"""
    FAST - Copyright (c) 2016 ONERA ISAE
"""

#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import openmdao.api as om

from fastoad.models.geometry.compute_aero_center import ComputeAeroCenter
from fastoad.models.geometry.geom_components import ComputeTotalArea
from fastoad.models.geometry.geom_components.fuselage.compute_fuselage import (
    ComputeFuselageGeometryBasic,
    ComputeFuselageGeometryCabinSizing,
)
from fastoad.models.geometry.geom_components.ht import ComputeHorizontalTailGeometry
from fastoad.models.geometry.geom_components.nacelle_pylons.compute_nacelle_pylons import (
    ComputeNacelleAndPylonsGeometry,
)
from fastoad.models.geometry.geom_components.vt import ComputeVerticalTailGeometry
from fastoad.models.geometry.geom_components.wing.compute_wing import ComputeWingGeometry
from fastoad.models.options import CABIN_SIZING_OPTION


class Geometry(om.Group):
    """
    Computes geometric characteristics of the (tube-wing) aircraft:
      - fuselage size is computed from payload requirements
      - wing dimensions are computed from global parameters (area, taper ratio...)
      - tail planes are dimensioned from HQ requirements

    This module also computes centers of gravity and static margin
    """

    def initialize(self):
        self.options.declare(CABIN_SIZING_OPTION, types=float, default=1.0)

    def setup(self):

        if self.options[CABIN_SIZING_OPTION] == 1.0:
            self.add_subsystem(
                "compute_fuselage", ComputeFuselageGeometryCabinSizing(), promotes=["*"]
            )
        else:
            self.add_subsystem("compute_fuselage", ComputeFuselageGeometryBasic(), promotes=["*"])

        self.add_subsystem("compute_wing", ComputeWingGeometry(), promotes=["*"])
        self.add_subsystem(
            "compute_engine_nacelle", ComputeNacelleAndPylonsGeometry(), promotes=["*"]
        )
        self.add_subsystem("compute_ht", ComputeHorizontalTailGeometry(), promotes=["*"])
        self.add_subsystem("compute_vt", ComputeVerticalTailGeometry(), promotes=["*"])
        self.add_subsystem("compute_total_area", ComputeTotalArea(), promotes=["*"])
        self.add_subsystem("compute_aero_center", ComputeAeroCenter(), promotes=["*"])



"""
    FAST - Copyright (c) 2016 ONERA ISAE
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from fastoad.models.aerodynamics.constants import POLAR_POINT_COUNT
from openmdao.core.explicitcomponent import ExplicitComponent


class Cd0Total(ExplicitComponent):
    def initialize(self):
        self.options.declare("low_speed_aero", default=False, types=bool)

    def setup(self):
        self.low_speed_aero = self.options["low_speed_aero"]

        self.add_input("data:geometry:aircraft:wetted_area", val=np.nan, units="m**2")

        nans_array = np.full(POLAR_POINT_COUNT, np.nan)
        if self.low_speed_aero:
            self.add_input("cd0_wing_low_speed", val=nans_array)
            self.add_input("cd0_fuselage_low_speed", val=nans_array)
            self.add_input("cd0_ht_low_speed", val=np.nan)
            self.add_input("cd0_vt_low_speed", val=np.nan)
            self.add_input("cd0_nacelle_low_speed", val=np.nan)
            self.add_input("cd0_pylon_low_speed", val=np.nan)
            self.add_output("cd0_total_low_speed", shape=POLAR_POINT_COUNT)
        else:
            self.add_input("data:aerodynamics:wing:cruise:CD0", val=nans_array)
            self.add_input("data:aerodynamics:fuselage:cruise:CD0", val=nans_array)
            self.add_input("data:aerodynamics:horizontal_tail:cruise:CD0", val=np.nan)
            self.add_input("data:aerodynamics:vertical_tail:cruise:CD0", val=np.nan)
            self.add_input("data:aerodynamics:nacelles:cruise:CD0", val=np.nan)
            self.add_input("data:aerodynamics:pylons:cruise:CD0", val=np.nan)
            self.add_output("data:aerodynamics:aircraft:cruise:CD0", shape=POLAR_POINT_COUNT)

        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):
        wet_area_total = inputs["data:geometry:aircraft:wetted_area"]
        if self.low_speed_aero:
            cd0_wing = inputs["cd0_wing_low_speed"]
            cd0_fus = inputs["cd0_fuselage_low_speed"]
            cd0_ht = inputs["cd0_ht_low_speed"]
            cd0_vt = inputs["cd0_vt_low_speed"]
            cd0_nac = inputs["cd0_nacelle_low_speed"]
            cd0_pylon = inputs["cd0_pylon_low_speed"]
        else:
            cd0_wing = inputs["data:aerodynamics:wing:cruise:CD0"]
            cd0_fus = inputs["data:aerodynamics:fuselage:cruise:CD0"]
            cd0_ht = inputs["data:aerodynamics:horizontal_tail:cruise:CD0"]
            cd0_vt = inputs["data:aerodynamics:vertical_tail:cruise:CD0"]
            cd0_nac = inputs["data:aerodynamics:nacelles:cruise:CD0"]
            cd0_pylon = inputs["data:aerodynamics:pylons:cruise:CD0"]

        k_techno = 1
        k_parasite = (
            -2.39 * pow(10, -12) * wet_area_total ** 3
            + 2.58 * pow(10, -8) * wet_area_total ** 2
            - 0.89 * pow(10, -4) * wet_area_total
            + 0.163
        )

        cd0_total_hs = cd0_wing + cd0_fus + cd0_ht + cd0_vt + cd0_nac + cd0_pylon
        cd0_total = cd0_total_hs * (1.0 + k_parasite * k_techno)

        if self.low_speed_aero:
            outputs["cd0_total_low_speed"] = cd0_total
        else:
            outputs["data:aerodynamics:aircraft:cruise:CD0"] = cd0_total



"""
    FAST - Copyright (c) 2016 ONERA ISAE
"""

#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from fastoad.models.aerodynamics.constants import POLAR_POINT_COUNT
from openmdao.core.explicitcomponent import ExplicitComponent


class Cd0Wing(ExplicitComponent):
    def initialize(self):
        self.options.declare("low_speed_aero", default=False, types=bool)

    def setup(self):
        self.low_speed_aero = self.options["low_speed_aero"]

        nans_array = np.full(POLAR_POINT_COUNT, np.nan)
        if self.low_speed_aero:
            self.add_input("reynolds_low_speed", val=np.nan)
            self.add_input("cl_low_speed", val=nans_array)
            self.add_input("Mach_low_speed", val=np.nan)
            self.add_output("cd0_wing_low_speed", shape=POLAR_POINT_COUNT)
        else:
            self.add_input("data:aerodynamics:wing:cruise:reynolds", val=np.nan)
            self.add_input("data:aerodynamics:aircraft:cruise:CL", val=nans_array)
            self.add_input("data:TLAR:cruise_mach", val=np.nan)
            self.add_output("data:aerodynamics:wing:cruise:CD0", shape=POLAR_POINT_COUNT)

        self.add_input("data:geometry:wing:area", val=np.nan, units="m**2")
        self.add_input("data:geometry:wing:thickness_ratio", val=np.nan)
        self.add_input("data:geometry:wing:wetted_area", val=np.nan, units="m**2")
        self.add_input("data:geometry:wing:MAC:length", val=np.nan, units="m")
        self.add_input("data:geometry:wing:sweep_25", val=np.nan, units="deg")

        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):
        wing_area = inputs["data:geometry:wing:area"]
        wet_area_wing = inputs["data:geometry:wing:wetted_area"]
        el_aero = inputs["data:geometry:wing:thickness_ratio"]
        sweep_25 = inputs["data:geometry:wing:sweep_25"]
        l0_wing = inputs["data:geometry:wing:MAC:length"]
        if self.low_speed_aero:
            cl = inputs["cl_low_speed"]
            mach = inputs["Mach_low_speed"]
            re_hs = inputs["reynolds_low_speed"]
        else:
            cl = inputs["data:aerodynamics:aircraft:cruise:CL"]
            mach = inputs["data:TLAR:cruise_mach"]
            re_hs = inputs["data:aerodynamics:wing:cruise:reynolds"]

        ki_arrow_cd0 = 0.04
        # Friction coefficients
        cf_wing_hs = 0.455 / ((1 + 0.144 * mach ** 2) ** 0.65 * (np.log10(re_hs * l0_wing)) ** 2.58)

        # cd0 wing
        # factor of relative thickness
        ke_cd0_wing = 4.688 * el_aero ** 2 + 3.146 * el_aero
        k_phi_cd0_wing = 1 - 0.000178 * sweep_25 ** 2 - 0.0065 * sweep_25

        kc_cd0_wing = (
            2.859 * (cl / np.cos(np.radians(sweep_25)) ** 2) ** 3
            - 1.849 * (cl / np.cos(np.radians(sweep_25)) ** 2) ** 2
            + 0.382 * (cl / np.cos(np.radians(sweep_25)) ** 2)
            + 0.06
        )  # sweep factor

        cd0_wing = (
            ((ke_cd0_wing + kc_cd0_wing) * k_phi_cd0_wing + ki_arrow_cd0 + 1)
            * cf_wing_hs
            * wet_area_wing
            / wing_area
        )

        if self.low_speed_aero:
            outputs["cd0_wing_low_speed"] = cd0_wing
        else:
            outputs["data:aerodynamics:wing:cruise:CD0"] = cd0_wing



"""
Tests basic XML serializer for OpenMDAO variables
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os.path as pth
from shutil import rmtree

import numpy as np
import pytest
from fastoad.io import VariableIO
from fastoad.io.xml import VariableXmlStandardFormatter
from fastoad.openmdao.variables import VariableList
from lxml import etree
from numpy.testing import assert_allclose

from ..exceptions import FastXPathEvalError

DATA_FOLDER_PATH = pth.join(pth.dirname(__file__), "data")
RESULTS_FOLDER_PATH = pth.join(
    pth.dirname(__file__), "results", pth.splitext(pth.basename(__file__))[0]
)


@pytest.fixture(scope="module")
def cleanup():
    rmtree(RESULTS_FOLDER_PATH, ignore_errors=True)


def _check_basic_vars(vars: VariableList):
    """ Checks that provided IndepVarComp instance matches content of data/basic.xml file """

    # Using pytest.approx for numerical reason, but also because it works even if sequence types
    # are different (lists, tuples, numpy arrays)
    assert_allclose(780.3, vars["geometry:total_surface"].value)
    assert vars["geometry:total_surface"].units == "m**2"

    assert_allclose(42, vars["geometry:wing:span"].value)
    assert vars["geometry:wing:span"].units == "m"

    assert_allclose(9.8, vars["geometry:wing:aspect_ratio"].value)
    assert vars["geometry:wing:aspect_ratio"].units is None

    assert_allclose(40.0, vars["geometry:fuselage:length"].value)
    assert vars["geometry:fuselage:length"].units == "m"

    assert_allclose(-42.0, vars["constants"].value)
    assert vars["constants"].units is None

    assert_allclose([1.0, 2.0, 3.0], vars["constants:k1"].value)
    assert vars["constants:k1"].units == "kg"

    assert_allclose([10.0, 20.0], vars["constants:k2"].value)
    assert vars["constants:k2"].units is None

    assert_allclose([100.0, 200.0, 300.0, 400.0], vars["constants:k3"].value)
    assert vars["constants:k3"].units == "m/s"

    assert_allclose([-1, -2, -3], vars["constants:k4"].value)
    assert vars["constants:k4"].units is None

    assert_allclose([100, 200, 400, 500, 600], vars["constants:k5"].value)
    assert vars["constants:k5"].units is None

    assert_allclose([[1e2, 3.4e5], [5.4e3, 2.1]], vars["constants:k8"].value)
    assert vars["constants:k8"].units is None

    assert len(vars) == 11


def test_basic_xml_read_and_write_from_vars(cleanup):
    """
    Tests the creation of an XML file from a VariableList instance
    """
    result_folder = pth.join(RESULTS_FOLDER_PATH, "basic_xml")

    # Check write hand-made component
    vars = VariableList()
    vars["geometry/total_surface"] = {"value": [780.3], "units": "m**2"}
    vars["geometry/wing/span"] = {"value": 42.0, "units": "m", "description": "span of the wing"}
    vars["geometry/wing/aspect_ratio"] = {"value": [9.8]}
    vars["geometry/fuselage/length"] = {"value": 40.0, "units": "m"}
    vars["constants"] = {"value": [-42.0], "description": "the answer"}
    vars["constants/k1"] = {"value": [1.0, 2.0, 3.0], "units": "kg"}
    vars["constants/k2"] = {"value": [10.0, 20.0], "description": "Geronimo!"}
    vars["constants/k3"] = {"value": np.array([100.0, 200.0, 300.0, 400.0]), "units": "m/s"}
    vars["constants/k4"] = {"value": [-1.0, -2.0, -3.0]}
    vars["constants/k5"] = {"value": [100.0, 200.0, 400.0, 500.0, 600.0]}
    vars["constants/k8"] = {"value": [[1e2, 3.4e5], [5.4e3, 2.1]]}

    # Try writing with non-existing folder
    assert not pth.exists(result_folder)
    filename = pth.join(result_folder, "handmade.xml")
    xml_write = VariableIO(filename, formatter=VariableXmlStandardFormatter())
    xml_write.path_separator = "/"
    xml_write.write(vars)

    # check (read another IndepVarComp instance from  xml)
    xml_check = VariableIO(filename, formatter=VariableXmlStandardFormatter())
    xml_check.path_separator = ":"
    new_vars = xml_check.read()
    _check_basic_vars(new_vars)

    # Check reading hand-made XML (with some format twists)
    filename = pth.join(DATA_FOLDER_PATH, "basic.xml")
    xml_read = VariableIO(filename, formatter=VariableXmlStandardFormatter())
    xml_read.path_separator = ":"
    vars = xml_read.read()
    _check_basic_vars(vars)

    # write it (with existing destination folder)
    new_filename = pth.join(result_folder, "basic.xml")
    xml_write = VariableIO(new_filename, formatter=VariableXmlStandardFormatter())
    xml_write.path_separator = ":"
    xml_write.write(vars)

    # check (read another IndepVarComp instance from new xml)
    xml_check = VariableIO(new_filename, formatter=VariableXmlStandardFormatter())
    xml_check.path_separator = ":"
    new_vars = xml_check.read()
    _check_basic_vars(new_vars)

    # try to write with bad separator
    xml_write.formatter.path_separator = "/"
    with pytest.raises(FastXPathEvalError):
        xml_write.write(vars)


def test_basic_xml_partial_read_and_write_from_vars(cleanup):
    """
    Tests the creation of an XML file from an IndepVarComp instance with only and ignore options
    """
    result_folder = pth.join(RESULTS_FOLDER_PATH, "basic_partial_xml")

    # Read full IndepVarComp
    filename = pth.join(DATA_FOLDER_PATH, "basic.xml")
    xml_read = VariableIO(filename, formatter=VariableXmlStandardFormatter())
    vars = xml_read.read(ignore=["does_not_exist"])
    _check_basic_vars(vars)

    # Add something to ignore and write it
    vars["should_be_ignored:pointless"] = {"value": 0.0}
    vars["should_also_be_ignored"] = {"value": -10.0}

    badvar_filename = pth.join(result_folder, "with_bad_var.xml")
    xml_write = VariableIO(badvar_filename, formatter=VariableXmlStandardFormatter())
    xml_write.write(vars, ignore=["does_not_exist"])  # Check with non-existent var in ignore list

    tree = etree.parse(badvar_filename)
    assert float(tree.xpath("should_be_ignored/pointless")[0].text.strip()) == 0.0
    assert float(tree.xpath("should_also_be_ignored")[0].text.strip()) == -10.0

    # Check partial reading with 'ignore'
    xml_read = VariableIO(badvar_filename, formatter=VariableXmlStandardFormatter())
    new_vars = xml_read.read(ignore=["should_be_ignored:pointless", "should_also_be_ignored"])
    _check_basic_vars(new_vars)

    # Check partial reading with 'only'
    ok_vars = [
        "geometry:total_surface",
        "geometry:wing:span",
        "geometry:wing:aspect_ratio",
        "geometry:fuselage:length",
        "constants",
        "constants:k1",
        "constants:k2",
        "constants:k3",
        "constants:k4",
        "constants:k5",
        "constants:k8",
    ]
    new_vars2 = xml_read.read(only=ok_vars)
    _check_basic_vars(new_vars2)

    # Check partial writing with 'ignore'
    varok_filename = pth.join(result_folder, "with_bad_var.xml")
    xml_write = VariableIO(varok_filename, formatter=VariableXmlStandardFormatter())
    xml_write.write(vars, ignore=["should_be_ignored:pointless", "should_also_be_ignored"])

    xml_read = VariableIO(varok_filename, formatter=VariableXmlStandardFormatter())
    new_vars = xml_read.read()
    _check_basic_vars(new_vars)

    # Check partial writing with 'only'
    varok2_filename = pth.join(result_folder, "with_bad_var.xml")
    xml_write = VariableIO(varok2_filename, formatter=VariableXmlStandardFormatter())
    xml_write.write(vars, only=ok_vars)

    xml_read = VariableIO(varok2_filename, formatter=VariableXmlStandardFormatter())
    new_vars = xml_read.read()
    _check_basic_vars(new_vars)



"""
    Estimation of global center of gravity
"""

#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from fastoad.models.weight.cg.cg_components import ComputeCGLoadCase1
from fastoad.models.weight.cg.cg_components.compute_cg_loadcase2 import ComputeCGLoadCase2
from fastoad.models.weight.cg.cg_components.compute_cg_loadcase3 import ComputeCGLoadCase3
from fastoad.models.weight.cg.cg_components.compute_cg_loadcase4 import ComputeCGLoadCase4
from fastoad.models.weight.cg.cg_components.compute_cg_ratio_aft import ComputeCGRatioAft
from fastoad.models.weight.cg.cg_components.compute_max_cg_ratio import ComputeMaxCGratio

from openmdao.api import Group


class ComputeGlobalCG(Group):
    # TODO: Document equations. Cite sources
    """ Global center of gravity estimation """

    def setup(self):
        self.add_subsystem("cg_ratio_aft", ComputeCGRatioAft(), promotes=["*"])
        self.add_subsystem("cg_ratio_lc1", ComputeCGLoadCase1(), promotes=["*"])
        self.add_subsystem("cg_ratio_lc2", ComputeCGLoadCase2(), promotes=["*"])
        self.add_subsystem("cg_ratio_lc3", ComputeCGLoadCase3(), promotes=["*"])
        self.add_subsystem("cg_ratio_lc4", ComputeCGLoadCase4(), promotes=["*"])
        self.add_subsystem("cg_ratio_max", ComputeMaxCGratio(), promotes=["*"])



"""
Tests for basic API
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import os.path as pth
import shutil
from filecmp import cmp
from shutil import rmtree

import pytest

from .. import api
from ..api import SAMPLE_FILENAME
from ..exceptions import FastFileExistsError

DATA_FOLDER_PATH = pth.join(pth.dirname(__file__), "data")
RESULTS_FOLDER_PATH = pth.join(pth.dirname(__file__), "results")
CONFIGURATION_FILE_PATH = pth.join(DATA_FOLDER_PATH, "sellar.toml")


@pytest.fixture(scope="module")
def cleanup():
    rmtree(RESULTS_FOLDER_PATH, ignore_errors=True)


def test_generate_configuration_file(cleanup):
    configuration_file_path = pth.join(RESULTS_FOLDER_PATH, "new_process.toml")

    api.generate_configuration_file(configuration_file_path, False)
    # Generating again without forcing overwrite will make it fail
    with pytest.raises(FastFileExistsError):
        api.generate_configuration_file(configuration_file_path, False)
    api.generate_configuration_file(configuration_file_path, True)

    original_file = pth.join(pth.dirname(api.__file__), "resources", SAMPLE_FILENAME)
    assert cmp(configuration_file_path, original_file)


def test_generate_inputs(cleanup):
    api.generate_inputs(CONFIGURATION_FILE_PATH, overwrite=False)
    # Generating again without forcing overwrite will make it fail
    with pytest.raises(FastFileExistsError):
        api.generate_inputs(CONFIGURATION_FILE_PATH, overwrite=False)
    api.generate_inputs(
        CONFIGURATION_FILE_PATH, pth.join(DATA_FOLDER_PATH, "inputs.xml"), overwrite=True
    )
    assert pth.exists(pth.join(RESULTS_FOLDER_PATH, "inputs.xml"))


def test_list_systems(cleanup):
    # Run with stdout output (no test)
    api.list_systems()
    api.list_systems(CONFIGURATION_FILE_PATH)

    # Run with file output (test file existence)
    out_file = pth.join(RESULTS_FOLDER_PATH, "list_systems.txt")
    assert not pth.exists(out_file)
    api.list_systems(CONFIGURATION_FILE_PATH, out_file)
    assert pth.exists(out_file)


def test_list_variables(cleanup):
    # Run with stdout output (no test)
    api.list_variables(CONFIGURATION_FILE_PATH)

    # Run with file output (test file existence)
    out_file = pth.join(RESULTS_FOLDER_PATH, "list_outputs.txt")
    assert not pth.exists(out_file)
    api.list_variables(CONFIGURATION_FILE_PATH, out_file)
    assert pth.exists(out_file)


def test_write_n2(cleanup):
    default_n2_file_path = pth.join(DATA_FOLDER_PATH, "n2.html")
    api.write_n2(CONFIGURATION_FILE_PATH)
    assert pth.exists(default_n2_file_path)
    os.remove(default_n2_file_path)

    n2_file_path = pth.join(RESULTS_FOLDER_PATH, "other_n2.html")
    api.write_n2(CONFIGURATION_FILE_PATH, n2_file_path)
    # Running again without forcing overwrite of outputs will make it fail
    with pytest.raises(FastFileExistsError):
        api.write_n2(CONFIGURATION_FILE_PATH, n2_file_path, False)
    api.write_n2(CONFIGURATION_FILE_PATH, n2_file_path, True)
    assert pth.exists(n2_file_path)


def test_evaluate_problem(cleanup):
    api.generate_inputs(
        CONFIGURATION_FILE_PATH, pth.join(DATA_FOLDER_PATH, "inputs.xml"), overwrite=True
    )
    api.evaluate_problem(CONFIGURATION_FILE_PATH, False)
    # Running again without forcing overwrite of outputs will make it fail
    with pytest.raises(FastFileExistsError):
        api.evaluate_problem(CONFIGURATION_FILE_PATH, False)
    problem = api.evaluate_problem(CONFIGURATION_FILE_PATH, True)
    assert problem["f"] == pytest.approx(32.56910089, abs=1e-8)

    # Move output file because it will be overwritten by the optim test
    shutil.move(
        pth.join(RESULTS_FOLDER_PATH, "outputs.xml"),
        pth.join(RESULTS_FOLDER_PATH, "outputs_eval.xml"),
    )


def test_optimize_problem(cleanup):
    api.generate_inputs(
        CONFIGURATION_FILE_PATH, pth.join(DATA_FOLDER_PATH, "inputs.xml"), overwrite=True
    )
    api.optimize_problem(CONFIGURATION_FILE_PATH, False)
    # Running again without forcing overwrite of outputs will make it fail
    with pytest.raises(FastFileExistsError):
        api.optimize_problem(CONFIGURATION_FILE_PATH, False)
    problem = api.optimize_problem(CONFIGURATION_FILE_PATH, True)

    assert problem["f"] == pytest.approx(3.18339395, abs=1e-8)


def test_optimization_viewer(cleanup):
    api.generate_inputs(
        CONFIGURATION_FILE_PATH, pth.join(DATA_FOLDER_PATH, "inputs.xml"), overwrite=True
    )

    # Before a run
    api.optimization_viewer(CONFIGURATION_FILE_PATH)

    api.optimize_problem(CONFIGURATION_FILE_PATH, True)

    # After a run
    api.optimization_viewer(CONFIGURATION_FILE_PATH)



"""
Test module for aerodynamics groups
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os.path as pth
from platform import system

import pytest
from fastoad.io import VariableIO
from pytest import approx

from tests.testing_utilities import run_system
from tests.xfoil_exe.get_xfoil import get_xfoil_path
from ..aerodynamics_high_speed import AerodynamicsHighSpeed
from ..aerodynamics_landing import AerodynamicsLanding
from ..aerodynamics_low_speed import AerodynamicsLowSpeed

xfoil_path = None if system() == "Windows" else get_xfoil_path()


def get_indep_var_comp(var_names):
    """ Reads required input data and returns an IndepVarcomp() instance"""
    reader = VariableIO(pth.join(pth.dirname(__file__), "data", "aerodynamics_inputs.xml"))
    reader.path_separator = ":"
    ivc = reader.read(only=var_names).to_ivc()
    return ivc


@pytest.mark.skipif(
    system() != "Windows" and xfoil_path is None, reason="No XFOIL executable available"
)
def test_aerodynamics_landing_with_xfoil():
    """ Tests AerodynamicsHighSpeed """
    input_list = [
        "data:TLAR:approach_speed",
        "data:mission:sizing:landing:flap_angle",
        "data:mission:sizing:landing:slat_angle",
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:thickness_ratio",
        "data:geometry:wing:sweep_25",
        "data:geometry:wing:sweep_0",
        "data:geometry:wing:sweep_100_outer",
        "data:geometry:flap:chord_ratio",
        "data:geometry:flap:span_ratio",
        "data:geometry:slat:chord_ratio",
        "data:geometry:slat:span_ratio",
        "xfoil:mach",
        "tuning:aerodynamics:aircraft:landing:CL_max:landing_gear_effect:k",
    ]

    ivc = get_indep_var_comp(input_list)
    problem = run_system(AerodynamicsLanding(xfoil_exe_path=xfoil_path), ivc)
    # Reference values are for Windows XFOIL version, but as results can be slightly different on other
    # platforms, tolerance is extended to 1e-2
    assert problem["data:aerodynamics:aircraft:landing:CL_max_clean"] == approx(1.59359, abs=1e-2)
    assert problem["data:aerodynamics:aircraft:landing:CL_max"] == approx(2.82178, abs=1e-2)


def test_aerodynamics_landing_without_xfoil():
    """ Tests AerodynamicsHighSpeed """
    input_list = [
        "data:TLAR:approach_speed",
        "data:mission:sizing:landing:flap_angle",
        "data:mission:sizing:landing:slat_angle",
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:thickness_ratio",
        "data:geometry:wing:sweep_25",
        "data:geometry:wing:sweep_0",
        "data:geometry:wing:sweep_100_outer",
        "data:geometry:flap:chord_ratio",
        "data:geometry:flap:span_ratio",
        "data:geometry:slat:chord_ratio",
        "data:geometry:slat:span_ratio",
        "xfoil:mach",
        "tuning:aerodynamics:aircraft:landing:CL_max:landing_gear_effect:k",
        "data:aerodynamics:aircraft:landing:CL_max_clean_2D",
    ]

    ivc = get_indep_var_comp(input_list)
    problem = run_system(AerodynamicsLanding(use_xfoil=False), ivc)
    assert problem["data:aerodynamics:aircraft:landing:CL_max_clean"] == approx(1.54978, abs=1e-5)
    assert problem["data:aerodynamics:aircraft:landing:CL_max"] == approx(2.77798, abs=1e-5)


def test_aerodynamics_high_speed():
    """ Tests AerodynamicsHighSpeed """
    input_list = [
        "data:geometry:propulsion:engine:count",
        "data:geometry:propulsion:fan:length",
        "data:geometry:flap:chord_ratio",
        "data:geometry:flap:span_ratio",
        "data:geometry:fuselage:maximum_height",
        "data:geometry:fuselage:length",
        "data:geometry:fuselage:wetted_area",
        "data:geometry:fuselage:maximum_width",
        "data:geometry:horizontal_tail:MAC:length",
        "data:geometry:horizontal_tail:sweep_25",
        "data:geometry:horizontal_tail:thickness_ratio",
        "data:geometry:horizontal_tail:wetted_area",
        "data:geometry:propulsion:nacelle:length",
        "data:geometry:propulsion:nacelle:wetted_area",
        "data:geometry:propulsion:pylon:length",
        "data:geometry:propulsion:pylon:wetted_area",
        "data:geometry:aircraft:wetted_area",
        "data:geometry:slat:chord_ratio",
        "data:geometry:slat:span_ratio",
        "data:geometry:vertical_tail:MAC:length",
        "data:geometry:vertical_tail:sweep_25",
        "data:geometry:vertical_tail:thickness_ratio",
        "data:geometry:vertical_tail:wetted_area",
        "data:geometry:wing:area",
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:root:chord",
        "data:geometry:wing:tip:chord",
        "data:geometry:wing:span",
        "data:geometry:wing:sweep_0",
        "data:geometry:wing:sweep_100_outer",
        "data:geometry:wing:sweep_25",
        "data:geometry:wing:thickness_ratio",
        "data:geometry:wing:wetted_area",
        "tuning:aerodynamics:aircraft:cruise:CD:k",
        "tuning:aerodynamics:aircraft:cruise:CL:k",
        "tuning:aerodynamics:aircraft:landing:CL_max:landing_gear_effect:k",
        "tuning:aerodynamics:aircraft:cruise:CD:winglet_effect:k",
        "tuning:aerodynamics:aircraft:cruise:CL:winglet_effect:k",
        "tuning:aerodynamics:aircraft:cruise:CD:offset",
        "tuning:aerodynamics:aircraft:cruise:CL:offset",
        "tuning:aerodynamics:aircraft:cruise:CD:winglet_effect:offset",
        "tuning:aerodynamics:aircraft:cruise:CL:winglet_effect:offset",
        "data:mission:sizing:cruise:altitude",
        "data:mission:sizing:landing:flap_angle",
        "data:mission:sizing:landing:slat_angle",
        "data:TLAR:cruise_mach",
        "data:TLAR:approach_speed",
    ]

    ivc = get_indep_var_comp(input_list)
    problem = run_system(AerodynamicsHighSpeed(), ivc)

    cd = problem["data:aerodynamics:aircraft:cruise:CD"]
    cl = problem["data:aerodynamics:aircraft:cruise:CL"]

    assert cd[cl == 0.0] == approx(0.02030, abs=1e-5)
    assert cd[cl == 0.2] == approx(0.02209, abs=1e-5)
    assert cd[cl == 0.42] == approx(0.02897, abs=1e-5)
    assert cd[cl == 0.85] == approx(0.11781, abs=1e-5)

    assert problem["data:aerodynamics:aircraft:cruise:optimal_CL"] == approx(0.54, abs=1e-3)
    assert problem["data:aerodynamics:aircraft:cruise:optimal_CD"] == approx(0.03550, abs=1e-5)


def test_aerodynamics_low_speed():
    """ Tests AerodynamicsLowSpeed """
    input_list = [
        "data:geometry:propulsion:engine:count",
        "data:geometry:propulsion:fan:length",
        "data:geometry:flap:chord_ratio",
        "data:geometry:flap:span_ratio",
        "data:geometry:fuselage:maximum_height",
        "data:geometry:fuselage:length",
        "data:geometry:fuselage:wetted_area",
        "data:geometry:fuselage:maximum_width",
        "data:geometry:horizontal_tail:MAC:length",
        "data:geometry:horizontal_tail:sweep_25",
        "data:geometry:horizontal_tail:thickness_ratio",
        "data:geometry:horizontal_tail:wetted_area",
        "data:geometry:propulsion:nacelle:length",
        "data:geometry:propulsion:nacelle:wetted_area",
        "data:geometry:propulsion:pylon:length",
        "data:geometry:propulsion:pylon:wetted_area",
        "data:geometry:aircraft:wetted_area",
        "data:geometry:slat:chord_ratio",
        "data:geometry:slat:span_ratio",
        "data:geometry:vertical_tail:MAC:length",
        "data:geometry:vertical_tail:sweep_25",
        "data:geometry:vertical_tail:thickness_ratio",
        "data:geometry:vertical_tail:wetted_area",
        "data:geometry:wing:area",
        "data:geometry:wing:aspect_ratio",
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:root:chord",
        "data:geometry:wing:tip:chord",
        "data:geometry:wing:span",
        "data:geometry:wing:sweep_0",
        "data:geometry:wing:sweep_100_outer",
        "data:geometry:wing:sweep_25",
        "data:geometry:wing:thickness_ratio",
        "data:geometry:wing:tip:thickness_ratio",
        "data:geometry:wing:wetted_area",
        "tuning:aerodynamics:aircraft:cruise:CD:k",
        "tuning:aerodynamics:aircraft:cruise:CL:k",
        "tuning:aerodynamics:aircraft:cruise:CD:winglet_effect:k",
        "tuning:aerodynamics:aircraft:cruise:CL:winglet_effect:k",
        "tuning:aerodynamics:aircraft:cruise:CD:offset",
        "tuning:aerodynamics:aircraft:cruise:CL:offset",
        "tuning:aerodynamics:aircraft:cruise:CD:winglet_effect:offset",
        "tuning:aerodynamics:aircraft:cruise:CL:winglet_effect:offset",
        "data:mission:sizing:landing:flap_angle",
        "data:mission:sizing:landing:slat_angle",
        "data:TLAR:approach_speed",
    ]

    ivc = get_indep_var_comp(input_list)
    problem = run_system(AerodynamicsLowSpeed(), ivc)

    cd = problem["aerodynamics:Cd_low_speed"]
    cl = problem["cl_low_speed"]

    assert cd[cl == 0.0] == approx(0.02173, abs=1e-5)
    assert cd[cl == 0.2] == approx(0.02339, abs=1e-5)
    assert cd[cl == 0.42] == approx(0.02974, abs=1e-5)
    assert cd[cl == 0.85] == approx(0.06010, abs=1e-5)



"""
    Estimation of geometry of horizontal tail
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import openmdao.api as om

from fastoad.models.geometry.geom_components.ht.components import ComputeHTChord
from fastoad.models.geometry.geom_components.ht.components import ComputeHTClalpha
from fastoad.models.geometry.geom_components.ht.components import ComputeHTMAC
from fastoad.models.geometry.geom_components.ht.components import ComputeHTSweep


class ComputeHorizontalTailGeometry(om.Group):
    """ Horizontal tail geometry estimation """

    def setup(self):
        self.add_subsystem("ht_chord", ComputeHTChord(), promotes=["*"])
        self.add_subsystem("ht_mac", ComputeHTMAC(), promotes=["*"])
        self.add_subsystem("ht_sweep", ComputeHTSweep(), promotes=["*"])
        self.add_subsystem("ht_cl_alpha", ComputeHTClalpha(), promotes=["*"])



"""
Test module for geometry functions of cg components
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os.path as pth

import pytest
from fastoad.io import VariableIO
from fastoad.models.weight.cg.cg_components import ComputeHTcg, ComputeVTcg, UpdateMLG

from tests.testing_utilities import run_system
from ..geom_components import ComputeTotalArea
from ..geom_components.fuselage import (
    ComputeFuselageGeometryBasic,
    ComputeFuselageGeometryCabinSizing,
)
from ..geom_components.fuselage.compute_cnbeta_fuselage import ComputeCnBetaFuselage
from ..geom_components.ht.components import (
    ComputeHTMAC,
    ComputeHTChord,
    ComputeHTClalpha,
    ComputeHTSweep,
)
from ..geom_components.nacelle_pylons.compute_nacelle_pylons import ComputeNacelleAndPylonsGeometry
from ..geom_components.vt.components import (
    ComputeVTMAC,
    ComputeVTChords,
    ComputeVTClalpha,
    ComputeVTSweep,
    ComputeVTDistance,
)
from ..geom_components.wing.components import (
    ComputeB50,
    ComputeCLalpha,
    ComputeL1AndL4Wing,
    ComputeL2AndL3Wing,
    ComputeMACWing,
    ComputeMFW,
    ComputeSweepWing,
    ComputeToCWing,
    ComputeWetAreaWing,
    ComputeXWing,
    ComputeYWing,
)


# pylint: disable=redefined-outer-name  # needed for pytest fixtures
@pytest.fixture(scope="module")
def input_xml() -> VariableIO:
    """
    :return: access to the sample xml data
    """
    # TODO: have more consistency in input data (no need for the whole geometry_inputs_full.xml)
    return VariableIO(pth.join(pth.dirname(__file__), "data", "geometry_inputs_full.xml"))


def test_compute_fuselage_cabin_sizing(input_xml):
    """ Tests computation of the fuselage with cabin sizing """

    input_list = [
        "data:geometry:cabin:seats:economical:width",
        "data:geometry:cabin:seats:economical:length",
        "data:geometry:cabin:seats:economical:count_by_row",
        "data:geometry:cabin:aisle_width",
        "data:geometry:cabin:exit_width",
        "data:TLAR:NPAX",
        "data:geometry:propulsion:engine:count",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    problem = run_system(ComputeFuselageGeometryCabinSizing(), input_vars)

    npax1 = problem["data:geometry:cabin:NPAX1"]
    assert npax1 == pytest.approx(157, abs=1)
    cg_systems_c6 = problem["data:weight:systems:flight_kit:CG:x"]
    assert cg_systems_c6 == pytest.approx(7.47, abs=1e-2)
    cg_furniture_d2 = problem["data:weight:furniture:passenger_seats:CG:x"]
    assert cg_furniture_d2 == pytest.approx(16.62, abs=1e-2)
    fuselage_length = problem["data:geometry:fuselage:length"]
    assert fuselage_length == pytest.approx(37.507, abs=1e-3)
    fuselage_width_max = problem["data:geometry:fuselage:maximum_width"]
    assert fuselage_width_max == pytest.approx(3.92, abs=1e-2)
    fuselage_height_max = problem["data:geometry:fuselage:maximum_height"]
    assert fuselage_height_max == pytest.approx(4.06, abs=1e-2)
    fuselage_lav = problem["data:geometry:fuselage:front_length"]
    assert fuselage_lav == pytest.approx(6.902, abs=1e-3)
    fuselage_lar = problem["data:geometry:fuselage:rear_length"]
    assert fuselage_lar == pytest.approx(14.616, abs=1e-3)
    fuselage_lpax = problem["data:geometry:fuselage:PAX_length"]
    assert fuselage_lpax == pytest.approx(22.87, abs=1e-2)
    fuselage_lcabin = problem["data:geometry:cabin:length"]
    assert fuselage_lcabin == pytest.approx(30.38, abs=1e-2)
    fuselage_wet_area = problem["data:geometry:fuselage:wetted_area"]
    assert fuselage_wet_area == pytest.approx(401.956, abs=1e-3)
    pnc = problem["data:geometry:cabin:crew_count:commercial"]
    assert pnc == pytest.approx(4, abs=1)


def test_compute_fuselage_basic(input_xml):
    """ Tests computation of the fuselage with no cabin sizing """

    input_list = [
        "data:geometry:cabin:NPAX1",
        "data:geometry:fuselage:length",
        "data:geometry:fuselage:maximum_width",
        "data:geometry:fuselage:maximum_height",
        "data:geometry:fuselage:front_length",
        "data:geometry:fuselage:rear_length",
        "data:geometry:fuselage:PAX_length",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    problem = run_system(ComputeFuselageGeometryBasic(), input_vars)

    cg_systems_c6 = problem["data:weight:systems:flight_kit:CG:x"]
    assert cg_systems_c6 == pytest.approx(9.19, abs=1e-2)
    cg_furniture_d2 = problem["data:weight:furniture:passenger_seats:CG:x"]
    assert cg_furniture_d2 == pytest.approx(14.91, abs=1e-2)
    fuselage_lcabin = problem["data:geometry:cabin:length"]
    assert fuselage_lcabin == pytest.approx(30.38, abs=1e-2)
    fuselage_wet_area = problem["data:geometry:fuselage:wetted_area"]
    assert fuselage_wet_area == pytest.approx(401.962, abs=1e-3)
    pnc = problem["data:geometry:cabin:crew_count:commercial"]
    assert pnc == pytest.approx(4, abs=1)


def test_compute_ht_cg(input_xml):
    """ Tests computation of the horizontal tail center of gravity """

    input_list = [
        "data:geometry:horizontal_tail:root:chord",
        "data:geometry:horizontal_tail:tip:chord",
        "data:geometry:horizontal_tail:MAC:at25percent:x:from_wingMAC25",
        "data:geometry:horizontal_tail:span",
        "data:geometry:wing:MAC:at25percent:x",
        "data:geometry:horizontal_tail:sweep_25",
        "data:geometry:horizontal_tail:MAC:length",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()
    input_vars.add_output("data:geometry:horizontal_tail:MAC:at25percent:x:local", 1.656, units="m")

    problem = run_system(ComputeHTcg(), input_vars)

    cg_a31 = problem["data:weight:airframe:horizontal_tail:CG:x"]
    assert cg_a31 == pytest.approx(34.58, abs=1e-2)


def test_compute_ht_mac(input_xml):
    """ Tests computation of the horizontal tail mac """

    input_list = [
        "data:geometry:horizontal_tail:root:chord",
        "data:geometry:horizontal_tail:tip:chord",
        "data:geometry:horizontal_tail:span",
        "data:geometry:horizontal_tail:sweep_25",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    problem = run_system(ComputeHTMAC(), input_vars)

    length = problem["data:geometry:horizontal_tail:MAC:length"]
    assert length == pytest.approx(3.141, abs=1e-3)
    ht_x0 = problem["data:geometry:horizontal_tail:MAC:at25percent:x:local"]
    assert ht_x0 == pytest.approx(1.656, abs=1e-3)
    ht_y0 = problem["data:geometry:horizontal_tail:MAC:y"]
    assert ht_y0 == pytest.approx(2.519, abs=1e-3)


def test_compute_ht_chord(input_xml):
    """ Tests computation of the horizontal tail chords """

    input_list = [
        "data:geometry:horizontal_tail:aspect_ratio",
        "data:geometry:horizontal_tail:area",
        "data:geometry:horizontal_tail:taper_ratio",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    problem = run_system(ComputeHTChord(), input_vars)

    span = problem["data:geometry:horizontal_tail:span"]
    assert span == pytest.approx(12.28, abs=1e-2)
    root_chord = problem["data:geometry:horizontal_tail:root:chord"]
    assert root_chord == pytest.approx(4.406, abs=1e-3)
    tip_chord = problem["data:geometry:horizontal_tail:tip:chord"]
    assert tip_chord == pytest.approx(1.322, abs=1e-3)


def test_compute_ht_cl(input_xml):
    """ Tests computation of the horizontal tail lift coefficient """

    input_list = [
        "data:geometry:horizontal_tail:aspect_ratio",
        "data:TLAR:cruise_mach",
        "data:geometry:horizontal_tail:sweep_25",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    problem = run_system(ComputeHTClalpha(), input_vars)

    cl_alpha = problem["data:aerodynamics:horizontal_tail:cruise:CL_alpha"]
    assert cl_alpha == pytest.approx(3.47, abs=1e-2)


def test_compute_ht_sweep(input_xml):
    """ Tests computation of the horizontal tail sweep """

    input_list = [
        "data:geometry:horizontal_tail:root:chord",
        "data:geometry:horizontal_tail:tip:chord",
        "data:geometry:horizontal_tail:span",
        "data:geometry:horizontal_tail:sweep_25",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    problem = run_system(ComputeHTSweep(), input_vars)

    sweep_0 = problem["data:geometry:horizontal_tail:sweep_0"]
    assert sweep_0 == pytest.approx(33.317, abs=1e-3)
    sweep_100 = problem["data:geometry:horizontal_tail:sweep_100"]
    assert sweep_100 == pytest.approx(8.81, abs=1e-2)


def test_compute_fuselage_cnbeta(input_xml):
    """ Tests computation of the yawing moment due to sideslip """

    input_list = [
        "data:geometry:fuselage:length",
        "data:geometry:fuselage:maximum_width",
        "data:geometry:fuselage:maximum_height",
        "data:geometry:fuselage:front_length",
        "data:geometry:fuselage:rear_length",
        "data:geometry:wing:area",
        "data:geometry:wing:span",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    component = ComputeCnBetaFuselage()

    problem = run_system(component, input_vars)

    cn_beta = problem["data:aerodynamics:fuselage:cruise:CnBeta"]
    assert cn_beta == pytest.approx(-0.117901, abs=1e-6)


def test_compute_vt_cg(input_xml):
    """ Tests computation of the vertical tail center of gravity """

    input_list = [
        "data:geometry:vertical_tail:root:chord",
        "data:geometry:vertical_tail:tip:chord",
        "data:geometry:vertical_tail:MAC:at25percent:x:from_wingMAC25",
        "data:geometry:vertical_tail:span",
        "data:geometry:wing:MAC:at25percent:x",
        "data:geometry:vertical_tail:sweep_25",
        "data:geometry:vertical_tail:MAC:length",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    input_vars.add_output("data:geometry:vertical_tail:MAC:at25percent:x:local", 2.321, units="m")

    component = ComputeVTcg()

    problem = run_system(component, input_vars)

    cg_a32 = problem["data:weight:airframe:vertical_tail:CG:x"]
    assert cg_a32 == pytest.approx(34.265, abs=1e-3)


def test_compute_vt_mac(input_xml):
    """ Tests computation of the vertical tail mac """

    input_list = [
        "data:geometry:vertical_tail:root:chord",
        "data:geometry:vertical_tail:tip:chord",
        "data:geometry:vertical_tail:span",
        "data:geometry:vertical_tail:sweep_25",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    component = ComputeVTMAC()

    problem = run_system(component, input_vars)

    length = problem["data:geometry:vertical_tail:MAC:length"]
    assert length == pytest.approx(4.161, abs=1e-3)
    vt_x0 = problem["data:geometry:vertical_tail:MAC:at25percent:x:local"]
    assert vt_x0 == pytest.approx(2.321, abs=1e-3)
    vt_z0 = problem["data:geometry:vertical_tail:MAC:z"]
    assert vt_z0 == pytest.approx(2.716, abs=1e-3)


def test_compute_vt_chords(input_xml):
    """ Tests computation of the vertical tail chords """

    input_list = [
        "data:geometry:vertical_tail:aspect_ratio",
        "data:geometry:vertical_tail:area",
        "data:geometry:vertical_tail:taper_ratio",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    component = ComputeVTChords()

    problem = run_system(component, input_vars)

    span = problem["data:geometry:vertical_tail:span"]
    assert span == pytest.approx(6.62, abs=1e-2)
    root_chord = problem["data:geometry:vertical_tail:root:chord"]
    assert root_chord == pytest.approx(5.837, abs=1e-3)
    tip_chord = problem["data:geometry:vertical_tail:tip:chord"]
    assert tip_chord == pytest.approx(1.751, abs=1e-3)


def test_compute_vt_sweep(input_xml):
    """ Tests computation of the vertical tail sweep """

    input_list = [
        "data:geometry:vertical_tail:root:chord",
        "data:geometry:vertical_tail:tip:chord",
        "data:geometry:vertical_tail:span",
        "data:geometry:vertical_tail:sweep_25",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    component = ComputeVTSweep()

    problem = run_system(component, input_vars)

    sweep_0 = problem["data:geometry:vertical_tail:sweep_0"]
    assert sweep_0 == pytest.approx(40.515, abs=1e-3)
    sweep_100 = problem["data:geometry:vertical_tail:sweep_100"]
    assert sweep_100 == pytest.approx(13.35, abs=1e-2)


def test_compute_vt_distance(input_xml):
    """ Tests computation of the vertical tail distance """

    input_list = [
        "data:geometry:fuselage:length",
        "data:geometry:wing:MAC:at25percent:x",
        "data:geometry:has_T_tail",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()
    problem = run_system(ComputeVTDistance(), input_vars)

    lp_vt = problem["data:geometry:vertical_tail:MAC:at25percent:x:from_wingMAC25"]
    assert lp_vt == pytest.approx(16.55, abs=1e-2)


def test_compute_vt_cl(input_xml):
    """ Tests computation of the vertical tail lift coefficient """

    input_list = [
        "data:TLAR:cruise_mach",
        "data:geometry:vertical_tail:aspect_ratio",
        "data:geometry:vertical_tail:sweep_25",
        "data:geometry:has_T_tail",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()
    problem = run_system(ComputeVTClalpha(), input_vars)

    cl_alpha = problem["data:aerodynamics:vertical_tail:cruise:CL_alpha"]
    assert cl_alpha == pytest.approx(2.55, abs=1e-2)


def test_geometry_wing_b50(input_xml):
    """ Tests computation of the wing B50 """

    input_list = [
        "data:geometry:wing:tip:leading_edge:x:local",
        "data:geometry:wing:root:y",
        "data:geometry:wing:tip:y",
        "data:geometry:wing:root:virtual_chord",
        "data:geometry:wing:tip:chord",
        "data:geometry:wing:span",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    component = ComputeB50()

    problem = run_system(component, input_vars)

    wing_b_50 = problem["data:geometry:wing:b_50"]
    assert wing_b_50 == pytest.approx(34.166, abs=1e-3)


def test_geometry_wing_cl_alpha(input_xml):
    """ Tests computation of the wing lift coefficient """

    input_list = [
        "data:TLAR:cruise_mach",
        "data:geometry:fuselage:maximum_width",
        "data:geometry:fuselage:maximum_height",
        "data:geometry:wing:area",
        "data:geometry:wing:root:chord",
        "data:geometry:wing:tip:chord",
        "data:geometry:wing:tip:thickness_ratio",
        "data:geometry:wing:sweep_25",
        "data:geometry:wing:aspect_ratio",
        "data:geometry:wing:span",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    component = ComputeCLalpha()

    problem = run_system(component, input_vars)

    cl_alpha = problem["data:aerodynamics:aircraft:cruise:CL_alpha"]
    assert cl_alpha == pytest.approx(6.49, abs=1e-2)


def test_geometry_wing_l1_l4(input_xml):
    """ Tests computation of the wing chords (l1 and l4) """

    input_list = [
        "data:geometry:wing:area",
        "data:geometry:wing:root:y",
        "data:geometry:wing:kink:y",
        "data:geometry:wing:span",
        "data:geometry:fuselage:maximum_width",
        "data:geometry:wing:taper_ratio",
        "data:geometry:wing:sweep_25",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    component = ComputeL1AndL4Wing()

    problem = run_system(component, input_vars)

    wing_l1 = problem["data:geometry:wing:root:virtual_chord"]
    assert wing_l1 == pytest.approx(4.953, abs=1e-3)
    wing_l4 = problem["data:geometry:wing:tip:chord"]
    assert wing_l4 == pytest.approx(1.882, abs=1e-3)


def test_geometry_wing_l2_l3(input_xml):
    """ Tests computation of the wing chords (l2 and l3) """

    input_list = [
        "data:geometry:wing:span",
        "data:geometry:fuselage:maximum_width",
        "data:geometry:wing:taper_ratio",
        "data:geometry:wing:sweep_25",
        "data:geometry:wing:root:virtual_chord",
        "data:geometry:wing:tip:chord",
        "data:geometry:wing:root:y",
        "data:geometry:wing:kink:y",
        "data:geometry:wing:tip:y",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    component = ComputeL2AndL3Wing()

    problem = run_system(component, input_vars)

    wing_l2 = problem["data:geometry:wing:root:chord"]
    assert wing_l2 == pytest.approx(6.26, abs=1e-2)
    wing_l3 = problem["data:geometry:wing:kink:chord"]
    assert wing_l3 == pytest.approx(3.985, abs=1e-3)


def test_geometry_wing_mac(input_xml):
    """ Tests computation of the wing mean aerodynamic chord """

    input_list = [
        "data:geometry:wing:area",
        "data:geometry:wing:kink:leading_edge:x:local",
        "data:geometry:wing:tip:leading_edge:x:local",
        "data:geometry:wing:root:y",
        "data:geometry:wing:kink:y",
        "data:geometry:wing:tip:y",
        "data:geometry:wing:root:chord",
        "data:geometry:wing:kink:chord",
        "data:geometry:wing:tip:chord",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    component = ComputeMACWing()

    problem = run_system(component, input_vars)

    wing_l0 = problem["data:geometry:wing:MAC:length"]
    assert wing_l0 == pytest.approx(4.457, abs=1e-3)
    wing_x0 = problem["data:geometry:wing:MAC:leading_edge:x:local"]
    assert wing_x0 == pytest.approx(2.361, abs=1e-3)
    wing_y0 = problem["data:geometry:wing:MAC:y"]
    assert wing_y0 == pytest.approx(6.293, abs=1e-3)


def test_geometry_wing_mfw(input_xml):
    """ Tests computation of the wing max fuel weight """

    input_list = [
        "data:geometry:wing:area",
        "data:geometry:wing:aspect_ratio",
        "data:geometry:wing:root:thickness_ratio",
        "data:geometry:wing:tip:thickness_ratio",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    component = ComputeMFW()

    problem = run_system(component, input_vars)

    mfw = problem["data:weight:aircraft:MFW"]
    assert mfw == pytest.approx(19284.7, abs=1e-1)


def test_geometry_wing_sweep(input_xml):
    """ Tests computation of the wing sweeps """

    input_list = [
        "data:geometry:wing:kink:leading_edge:x:local",
        "data:geometry:wing:tip:leading_edge:x:local",
        "data:geometry:wing:root:y",
        "data:geometry:wing:kink:y",
        "data:geometry:wing:tip:y",
        "data:geometry:wing:root:chord",
        "data:geometry:wing:kink:chord",
        "data:geometry:wing:tip:chord",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    component = ComputeSweepWing()

    problem = run_system(component, input_vars)

    sweep_0 = problem["data:geometry:wing:sweep_0"]
    assert sweep_0 == pytest.approx(27.55, abs=1e-2)
    sweep_100_inner = problem["data:geometry:wing:sweep_100_inner"]
    assert sweep_100_inner == pytest.approx(0.0, abs=1e-1)
    sweep_100_outer = problem["data:geometry:wing:sweep_100_outer"]
    assert sweep_100_outer == pytest.approx(16.7, abs=1e-1)


def test_geometry_wing_toc(input_xml):
    """ Tests computation of the wing ToC (Thickness of Chord) """

    input_list = ["data:TLAR:cruise_mach", "data:geometry:wing:sweep_25"]

    input_vars = input_xml.read(only=input_list).to_ivc()

    component = ComputeToCWing()

    problem = run_system(component, input_vars)

    toc_aero = problem["data:geometry:wing:thickness_ratio"]
    assert toc_aero == pytest.approx(0.128, abs=1e-3)
    toc_root = problem["data:geometry:wing:root:thickness_ratio"]
    assert toc_root == pytest.approx(0.159, abs=1e-3)
    toc_kink = problem["data:geometry:wing:kink:thickness_ratio"]
    assert toc_kink == pytest.approx(0.121, abs=1e-3)
    toc_tip = problem["data:geometry:wing:tip:thickness_ratio"]
    assert toc_tip == pytest.approx(0.11, abs=1e-2)


def test_geometry_wing_wet_area(input_xml):
    """ Tests computation of the wing wet area """

    input_list = [
        "data:geometry:wing:root:chord",
        "data:geometry:wing:root:y",
        "data:geometry:wing:area",
        "data:geometry:fuselage:maximum_width",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    component = ComputeWetAreaWing()

    problem = run_system(component, input_vars)

    area_pf = problem["data:geometry:wing:outer_area"]
    assert area_pf == pytest.approx(100.303, abs=1e-3)
    wet_area = problem["data:geometry:wing:wetted_area"]
    assert wet_area == pytest.approx(200.607, abs=1e-3)


def test_geometry_wing_x(input_xml):
    """ Tests computation of the wing Xs """

    input_list = [
        "data:geometry:wing:root:virtual_chord",
        "data:geometry:wing:kink:chord",
        "data:geometry:wing:tip:chord",
        "data:geometry:wing:root:y",
        "data:geometry:wing:kink:y",
        "data:geometry:wing:tip:y",
        "data:geometry:wing:sweep_25",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    component = ComputeXWing()

    problem = run_system(component, input_vars)

    wing_x3 = problem["data:geometry:wing:kink:leading_edge:x:local"]
    assert wing_x3 == pytest.approx(2.275, abs=1e-3)
    wing_x4 = problem["data:geometry:wing:tip:leading_edge:x:local"]
    assert wing_x4 == pytest.approx(7.222, abs=1e-3)


def test_geometry_wing_y(input_xml):
    """ Tests computation of the wing Ys """

    input_list = [
        "data:geometry:wing:aspect_ratio",
        "data:geometry:fuselage:maximum_width",
        "data:geometry:wing:area",
        "data:geometry:wing:kink:span_ratio",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    component = ComputeYWing()

    problem = run_system(component, input_vars)

    span = problem["data:geometry:wing:span"]
    assert span == pytest.approx(34.4, abs=1e-1)
    wing_y2 = problem["data:geometry:wing:root:y"]
    assert wing_y2 == pytest.approx(1.96, abs=1e-2)
    wing_y3 = problem["data:geometry:wing:kink:y"]
    assert wing_y3 == pytest.approx(6.88, abs=1e-2)
    wing_y4 = problem["data:geometry:wing:tip:y"]
    assert wing_y4 == pytest.approx(17.2, abs=1e-1)


def test_geometry_nacelle_pylons(input_xml):
    """ Tests computation of the nacelle and pylons component """

    input_list = [
        "data:propulsion:MTO_thrust",
        "data:geometry:propulsion:engine:y_ratio",
        "data:geometry:propulsion:layout",
        "data:geometry:wing:span",
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:MAC:leading_edge:x:local",
        "data:geometry:wing:root:chord",
        "data:geometry:wing:root:y",
        "data:geometry:wing:kink:chord",
        "data:geometry:wing:kink:y",
        "data:geometry:wing:kink:leading_edge:x:local",
        "data:geometry:wing:MAC:at25percent:x",
        "data:geometry:fuselage:length",
        "data:geometry:fuselage:maximum_width",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    component = ComputeNacelleAndPylonsGeometry()

    problem = run_system(component, input_vars)

    pylon_length = problem["data:geometry:propulsion:pylon:length"]
    assert pylon_length == pytest.approx(5.733, abs=1e-3)
    fan_length = problem["data:geometry:propulsion:fan:length"]
    assert fan_length == pytest.approx(3.127, abs=1e-3)
    nacelle_length = problem["data:geometry:propulsion:nacelle:length"]
    assert nacelle_length == pytest.approx(5.211, abs=1e-3)
    nacelle_dia = problem["data:geometry:propulsion:nacelle:diameter"]
    assert nacelle_dia == pytest.approx(2.172, abs=1e-3)
    lg_height = problem["data:geometry:landing_gear:height"]
    assert lg_height == pytest.approx(3.041, abs=1e-3)
    y_nacell = problem["data:geometry:propulsion:nacelle:y"]
    assert y_nacell == pytest.approx(5.373, abs=1e-3)
    pylon_wet_area = problem["data:geometry:propulsion:pylon:wetted_area"]
    assert pylon_wet_area == pytest.approx(7.563, abs=1e-3)
    nacelle_wet_area = problem["data:geometry:propulsion:nacelle:wetted_area"]
    assert nacelle_wet_area == pytest.approx(21.609, abs=1e-3)
    cg_b1 = problem["data:weight:propulsion:engine:CG:x"]
    assert cg_b1 == pytest.approx(13.5, abs=1e-1)


def test_geometry_total_area(input_xml):
    """ Tests computation of the total area """

    input_list = [
        "data:geometry:wing:wetted_area",
        "data:geometry:fuselage:wetted_area",
        "data:geometry:horizontal_tail:wetted_area",
        "data:geometry:vertical_tail:wetted_area",
        "data:geometry:propulsion:nacelle:wetted_area",
        "data:geometry:propulsion:pylon:wetted_area",
        "data:geometry:propulsion:engine:count",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    component = ComputeTotalArea()

    problem = run_system(component, input_vars)

    total_surface = problem["data:geometry:aircraft:wetted_area"]
    assert total_surface == pytest.approx(783.997, abs=1e-3)


def test_geometry_update_mlg(input_xml):
    """ Tests computation of the main landing gear """

    input_list = [
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:MAC:at25percent:x",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    input_vars.add_output("data:weight:aircraft:CG:aft:MAC_position", 0.364924)
    input_vars.add_output("data:weight:airframe:landing_gear:front:CG:x", 5.07)

    problem = run_system(UpdateMLG(), input_vars)

    cg_a51 = problem["data:weight:airframe:landing_gear:main:CG:x"]
    assert cg_a51 == pytest.approx(18.00, abs=1e-2)



"""
Test module for OpenMDAO checks
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging

import openmdao.api as om

from .sellar_example.disc1 import Disc1
from .sellar_example.disc2 import Disc2
from .sellar_example.functions import Functions
from .sellar_example.sellar import Sellar
from ..utils import get_unconnected_input_names

# Logger for this module
_LOGGER = logging.getLogger(__name__)


def test_get_unconnected_inputs():
    """ Tests get_unconnected_inputs() on a minimal problem  """
    # Check with Component as problem model -----------------------------------

    test_labels = []
    components = {}
    expected_mandatory_variables = {}
    expected_optional_variables = {}

    # Test with a problem with a single component
    test_label = "single_component"
    test_labels.append(test_label)
    components[test_label] = Disc1()
    expected_mandatory_variables[test_label] = ["x"]
    expected_optional_variables[test_label] = ["y2", "z"]

    # Test with a group problem with a single component (no promoted variable)
    test_label = "single_component_group"
    test_labels.append(test_label)
    group = om.Group()
    group.add_subsystem("disc1", Disc1())
    components[test_label] = group
    expected_mandatory_variables[test_label] = ["disc1.x"]
    expected_optional_variables[test_label] = ["disc1.y2", "disc1.z"]

    # Test with a group problem with one component and some input
    test_label = "component+some_input"
    test_labels.append(test_label)
    group = om.Group()
    group.add_subsystem("disc1", Disc1())
    ivc = om.IndepVarComp()
    ivc.add_output("y2", 1.0)
    group.add_subsystem("inputs", ivc)
    group.connect("inputs.y2", "disc1.y2")
    components[test_label] = group
    expected_mandatory_variables[test_label] = ["disc1.x"]
    expected_optional_variables[test_label] = ["disc1.z"]

    # Test with a group problem with Sellar components
    test_label = "Sellar_components"
    test_labels.append(test_label)
    group = om.Group()
    group.add_subsystem("disc1", Disc1(), promotes=["*"])
    group.add_subsystem("disc2", Disc2(), promotes=["*"])
    group.add_subsystem("functions", Functions(), promotes=["*"])
    components[test_label] = group
    expected_mandatory_variables[test_label] = ["disc1.x", "functions.z"]
    expected_optional_variables[test_label] = ["disc1.z", "disc2.z", "functions.x"]

    # Test with the fully set Sellar problem
    test_label = "Sellar"
    test_labels.append(test_label)
    components[test_label] = Sellar()
    expected_mandatory_variables[test_label] = []
    expected_optional_variables[test_label] = []

    for label in test_labels:
        _LOGGER.info("Testing %s -------------------" % label)
        problem = om.Problem(components[label])
        _test_problem(
            problem, expected_mandatory_variables[label], expected_optional_variables[label]
        )


def _test_problem(
    problem, expected_missing_mandatory_variables, expected_missing_optional_variables
):
    """ Tests get_unconnected_inputs for provided problem """

    # with logger provided
    mandatory, optional = get_unconnected_input_names(problem, logger=_LOGGER)
    assert sorted(mandatory) == sorted(expected_missing_mandatory_variables)
    assert sorted(optional) == sorted(expected_missing_optional_variables)

    # without providing logger
    mandatory, optional = get_unconnected_input_names(problem)
    assert sorted(mandatory) == sorted(expected_missing_mandatory_variables)
    assert sorted(optional) == sorted(expected_missing_optional_variables)



"""
Test module for variable_io_legacy.py
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os.path as pth
from shutil import rmtree

import pytest
from numpy.testing import assert_allclose

from .. import VariableLegacy1XmlFormatter
from ... import VariableIO

DATA_FOLDER_PATH = pth.join(pth.dirname(__file__), "data")
RESULTS_FOLDER_PATH = pth.join(
    pth.dirname(__file__), "results", pth.splitext(pth.basename(__file__))[0]
)


@pytest.fixture(scope="module")
def cleanup():
    rmtree(RESULTS_FOLDER_PATH, ignore_errors=True)


def test_legacy1(cleanup):
    """ Tests class OMLegacy1XmlIO """
    result_folder = pth.join(RESULTS_FOLDER_PATH, "legacy1_xml")

    # test read ---------------------------------------------------------------
    filename = pth.join(DATA_FOLDER_PATH, "CeRAS01_baseline.xml")

    xml_read = VariableIO(filename, formatter=VariableLegacy1XmlFormatter())
    var_list = xml_read.read()

    entry_count = len(var_list)
    # Entry count may vary according to changes in translation table, but
    # we check that enough values are actually read.
    assert entry_count > 400

    # Check some random fields
    assert_allclose(var_list["data:geometry:wing:MAC:at25percent:x"].value, 16.457)
    assert var_list["data:geometry:wing:MAC:at25percent:x"].units == "m"
    assert_allclose(var_list["data:TLAR:NPAX"].value, 150)
    assert var_list["data:TLAR:NPAX"].units is None
    assert_allclose(var_list["data:geometry:wing:wetted_area"].value, 200.607)
    assert var_list["data:geometry:wing:wetted_area"].units == "m**2"

    # test write ---------------------------------------------------------------
    new_filename = pth.join(result_folder, "CeRAS01_baseline.xml")
    xml_write = VariableIO(new_filename, formatter=VariableLegacy1XmlFormatter())
    xml_write.write(var_list)

    # check by reading without conversion table
    # -> this will give the actual number of entries in the file
    xml_check = VariableIO(new_filename)
    check_vars = xml_check.read()
    assert len(check_vars) == entry_count



"""
Module for testing VariableList.py
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List

import numpy as np
import openmdao.api as om
import pytest
from numpy.testing import assert_allclose

from .sellar_example.disc1 import Disc1
from .sellar_example.disc2 import Disc2
from .sellar_example.functions import Functions
from ..variables import VariableList, Variable


def test_variables():
    """ Tests features of Variable and VariableList class"""

    # Test description overloading
    x = Variable("test:test_variable", value=500)
    assert x.description == "for testing (do not remove, keep first)"

    # Initialization
    variables = VariableList()
    a_var = Variable("a", value=0.0)
    b_var = Variable("b", value=1.0)
    n_var = Variable("n", value=np.array(np.nan))
    variables["a"] = {"value": 0.0}  # Tests VariableList.__setitem__ with dict input
    variables.append(b_var)  # Tests VariableList.append()
    with pytest.raises(TypeError):
        variables["z"] = 5.0  # error when value is not a dict
    with pytest.raises(TypeError):
        variables[1] = 5.0  # error when value is not a Variable
    with pytest.raises(TypeError):
        variables.append(5.0)  # error when value is not a Variable

    variables.append(n_var)
    variables[2] = n_var  # same as line above
    del variables["n"]

    # Initialization from list
    variables2 = VariableList([a_var, b_var])
    assert variables == variables2

    # tests on Variable
    assert a_var.value == 0.0
    assert a_var.units is None
    assert a_var.description == ""

    assert n_var == Variable("n", value=np.array(np.nan))  # tests __eq__ with nan value

    #   __getitem___
    assert variables["a"] == a_var
    assert variables["b"] is b_var

    #   .names()
    assert list(variables.names()) == ["a", "b"]

    # Tests adding variable with existing name
    assert len(variables) == 2
    assert variables["a"].value == 0.0
    variables.append(Variable("a", value=5.0))
    assert len(variables) == 2
    assert variables["a"].value == 5.0
    variables["a"] = {"value": 42.0}
    assert variables["a"].value == 42.0

    # .update()
    assert len(variables) == 2
    assert list(variables.names()) == ["a", "b"]
    variables.update([n_var])  # does nothing
    assert len(variables) == 2
    assert list(variables.names()) == ["a", "b"]

    variables.update([n_var], add_variables=True)
    assert len(variables) == 3
    assert list(variables.names()) == ["a", "b", "n"]
    assert variables["a"].value == 42.0

    variables.update([Variable("a", value=-10.0), Variable("not_added", value=0.0)])
    assert len(variables) == 3
    assert list(variables.names()) == ["a", "b", "n"]
    assert variables["a"].value == -10.0


def test_ivc_from_to_variables():
    """
    Tests VariableList.to_ivc() and VariableList.from_ivc()
    """
    vars = VariableList()
    vars["a"] = {"value": 5}
    vars["b"] = {"value": 2.5, "units": "m"}
    vars["c"] = {"value": -3.2, "units": "kg/s", "desc": "some test"}

    ivc = vars.to_ivc()
    problem = om.Problem(ivc)
    problem.setup()
    assert problem["a"] == 5
    assert problem.get_val("b", units="cm") == 250
    assert problem.get_val("c", units="kg/ms") == -0.0032

    new_vars = VariableList.from_ivc(ivc)
    assert vars.names() == new_vars.names()
    for var, new_var in zip(vars, new_vars):
        assert var == new_var


def test_df_from_to_variables():
    """
    Tests VariableList.to_dataframe() and VariableList.from_dataframe()
    """
    vars = VariableList()
    vars["a"] = {"value": 5}
    vars["b"] = {"value": np.array([1.0, 2.0, 3.0]), "units": "m"}
    vars["c"] = {"value": [1.0, 2.0, 3.0], "units": "kg/s", "desc": "some test"}

    df = vars.to_dataframe()
    assert np.all(df["name"] == ["a", "b", "c"])
    assert np.all(df["value"] == [5, [1.0, 2.0, 3.0], [1.0, 2.0, 3.0]])
    assert np.all(df["units"].to_list() == [None, "m", "kg/s"])
    assert np.all(df["desc"].to_list() == ["", "", "some test"])

    new_vars = VariableList.from_dataframe(df)

    assert vars.names() == new_vars.names()
    for var, new_var in zip(vars, new_vars):
        assert var == new_var


def test_get_variables_from_system():
    def _test_and_check(
        component, use_inputs: bool, use_outputs: bool, expected_vars: List[Variable],
    ):
        vars = VariableList.from_system(component, use_inputs=use_inputs, use_outputs=use_outputs)

        # A comparison of sets will not work due to values not being strictly equal
        # (not enough decimals in expected values), so we do not use this:
        # assert set(vars) == set(expected_vars)
        # Simply comparing variable name, units and approximate value will do for here.
        sort_key = lambda v: v.name
        vars.sort(key=sort_key)
        expected_vars.sort(key=sort_key)
        for var, expected_var in zip(vars, expected_vars):
            assert var.name == expected_var.name
            assert_allclose(var.value, expected_var.value)
            assert var.units == expected_var.units

    # test Component -----------------------------------------------------------
    comp = Disc1()
    expected_input_vars = [
        Variable(name="x", value=np.array([np.nan]), units=None),
        Variable(name="y2", value=np.array([1.0]), units=None),
        Variable(name="z", value=np.array([5.0, 2.0]), units="m**2"),
    ]
    expected_output_vars = [Variable(name="y1", value=np.array([1.0]), units=None)]
    _test_and_check(comp, True, False, expected_input_vars)
    _test_and_check(comp, False, True, expected_output_vars)
    _test_and_check(comp, True, True, expected_input_vars + expected_output_vars)

    # test Group ---------------------------------------------------------------
    group = om.Group()
    group.add_subsystem("disc1", Disc1(), promotes=["*"])
    group.add_subsystem("disc2", Disc2(), promotes=["*"])

    expected_input_vars = [
        Variable(name="x", value=np.array([np.nan]), units=None),
        Variable(name="y1", value=np.array([1.0]), units=None),
        Variable(name="y2", value=np.array([1.0]), units=None),
        Variable(name="z", value=np.array([5.0, 2.0]), units="m**2"),
    ]
    expected_output_vars = [
        Variable(name="y1", value=np.array([1.0]), units=None),
        Variable(name="y2", value=np.array([1.0]), units=None),
    ]
    _test_and_check(group, True, False, expected_input_vars)
    _test_and_check(group, False, True, expected_output_vars)
    _test_and_check(group, True, True, expected_input_vars)


def test_get_variables_from_problem():
    def _test_and_check(
        problem: om.Problem,
        use_initial_values: bool,
        use_inputs: bool,
        use_outputs: bool,
        expected_vars: List[Variable],
    ):
        vars = VariableList.from_problem(
            problem, use_initial_values, use_inputs=use_inputs, use_outputs=use_outputs
        )

        # A comparison of sets will not work due to values not being strictly equal
        # (not enough decimals in expected values), so we do not use this:
        # assert set(vars) == set(expected_vars)
        # Simply comparing variable name, units and approximate value will do for here.
        sort_key = lambda v: v.name
        vars.sort(key=sort_key)
        expected_vars.sort(key=sort_key)
        for var, expected_var in zip(vars, expected_vars):
            assert var.name == expected_var.name
            assert_allclose(var.value, expected_var.value)
            assert var.units == expected_var.units

    # Check with an ExplicitComponent ------------------------------------------
    problem = om.Problem(Disc1())
    expected_input_vars = [
        Variable(name="x", value=np.array([np.nan]), units=None),
        Variable(name="y2", value=np.array([1.0]), units=None),
        Variable(name="z", value=np.array([5.0, 2.0]), units="m**2"),
    ]
    expected_output_vars = [Variable(name="y1", value=np.array([1.0]), units=None)]
    _test_and_check(problem, False, True, False, expected_input_vars)
    _test_and_check(problem, False, False, True, expected_output_vars)
    _test_and_check(problem, False, True, True, expected_input_vars + expected_output_vars)

    # Check with a Group -------------------------------------------------------
    group = om.Group()
    group.add_subsystem("disc1", Disc1(), promotes=["*"])
    group.add_subsystem("disc2", Disc2(), promotes=["*"])
    problem = om.Problem(group)

    # All variables are inputs somewhere
    expected_input_vars = [
        Variable(name="x", value=np.array([np.nan]), units=None),
        Variable(name="y1", value=np.array([1.0]), units=None),
        Variable(name="y2", value=np.array([1.0]), units=None),
        Variable(name="z", value=np.array([5.0, 2.0]), units="m**2"),
    ]
    expected_output_vars = [
        Variable(name="y1", value=np.array([1.0]), units=None),
        Variable(name="y2", value=np.array([1.0]), units=None),
    ]
    _test_and_check(problem, False, True, False, expected_input_vars)
    _test_and_check(problem, False, False, True, expected_output_vars)
    _test_and_check(problem, False, True, True, expected_input_vars)

    # Check with the whole Sellar problem --------------------------------------
    # WITH promotions, WITHOUT computation
    group = om.Group()
    indeps = group.add_subsystem("indeps", om.IndepVarComp(), promotes=["*"])
    indeps.add_output("x", 1.0, units="Pa")  # This setting of units will prevail in our output
    indeps.add_output("z", [5.0, 2.0], units="m**2")
    group.add_subsystem("disc1", Disc1(), promotes=["*"])
    group.add_subsystem("disc2", Disc2(), promotes=["*"])
    group.add_subsystem("functions", Functions(), promotes=["*"])
    group.nonlinear_solver = om.NonlinearBlockGS()
    problem = om.Problem(group)

    expected_input_vars = [
        Variable(name="x", value=np.array([np.nan]), units=None),
        Variable(name="z", value=np.array([5.0, 2.0]), units="m**2"),
        Variable(name="y1", value=np.array([1.0]), units=None),
        Variable(name="y2", value=np.array([1.0]), units=None),
    ]
    expected_output_vars = [
        Variable(name="x", value=np.array([1.0]), units="Pa"),
        Variable(name="z", value=np.array([5.0, 2.0]), units="m**2"),
        Variable(name="y1", value=np.array([1.0]), units=None),
        Variable(name="y2", value=np.array([1.0]), units=None),
        Variable(name="g1", value=np.array([1.0]), units=None),
        Variable(name="g2", value=np.array([1.0]), units=None),
        Variable(name="f", value=np.array([1.0]), units=None),
    ]
    _test_and_check(problem, True, True, False, expected_input_vars)
    _test_and_check(problem, True, False, True, expected_output_vars)

    # Check with the whole Sellar problem --------------------------------------
    # WITH promotions, WITH computation
    expected_computed_output_vars = [
        Variable(name="x", value=np.array([1.0]), units="Pa"),
        Variable(name="z", value=np.array([5.0, 2.0]), units="m**2"),
        Variable(name="y1", value=np.array([25.58830237]), units=None),
        Variable(name="y2", value=np.array([12.05848815]), units=None),
        Variable(name="f", value=np.array([28.58830817]), units=None),
        Variable(name="g1", value=np.array([-22.42830237]), units=None),
        Variable(name="g2", value=np.array([-11.94151185]), units=None),
    ]
    problem.setup()
    problem.run_model()
    _test_and_check(problem, True, True, False, expected_input_vars)
    _test_and_check(problem, True, False, True, expected_output_vars)
    _test_and_check(problem, False, False, True, expected_computed_output_vars)

    # Check with the whole Sellar problem --------------------------------------
    # WITHOUT promotions, WITHOUT computation
    group = om.Group()
    indeps = group.add_subsystem("indeps", om.IndepVarComp())
    indeps.add_output("x", 1.0, units="Pa")
    indeps.add_output("z", [5.0, 2.0], units="m**2")
    group.add_subsystem("disc2", Disc2())
    group.add_subsystem("disc1", Disc1())
    group.add_subsystem("functions", Functions())
    group.nonlinear_solver = om.NonlinearBlockGS()
    group.connect("indeps.x", "disc1.x")
    group.connect("indeps.x", "functions.x")
    group.connect("indeps.z", "disc1.z")
    group.connect("indeps.z", "disc2.z")
    group.connect("indeps.z", "functions.z")
    group.connect("disc1.y1", "disc2.y1")
    group.connect("disc1.y1", "functions.y1")
    group.connect("disc2.y2", "disc1.y2")
    group.connect("disc2.y2", "functions.y2")

    problem = om.Problem(group)

    expected_vars = [
        Variable(name="indeps.x", value=np.array([1.0]), units="Pa"),
        Variable(name="indeps.z", value=np.array([5.0, 2.0]), units="m**2"),
        Variable(name="disc1.x", value=np.array([np.nan]), units=None),
        Variable(name="disc1.z", value=np.array([5.0, 2.0]), units="m**2"),
        Variable(name="disc1.y1", value=np.array([1.0]), units=None),
        Variable(name="disc1.y2", value=np.array([1.0]), units=None),
        Variable(name="disc2.z", value=np.array([5.0, 2.0]), units="m**2"),
        Variable(name="disc2.y1", value=np.array([1.0]), units=None),
        Variable(name="disc2.y2", value=np.array([1.0]), units=None),
        Variable(name="functions.x", value=np.array([2]), units=None),
        Variable(name="functions.z", value=np.array([np.nan, np.nan]), units="m**2"),
        Variable(name="functions.y1", value=np.array([1.0]), units=None),
        Variable(name="functions.y2", value=np.array([1.0]), units=None),
        Variable(name="functions.g1", value=np.array([1.0]), units=None),
        Variable(name="functions.g2", value=np.array([1.0]), units=None),
        Variable(name="functions.f", value=np.array([1.0]), units=None),
    ]
    _test_and_check(problem, True, True, True, expected_vars)
    expected_computed_vars = [  # Here links are done, even without computations
        Variable(name="indeps.x", value=np.array([1.0]), units="Pa"),
        Variable(name="indeps.z", value=np.array([5.0, 2.0]), units="m**2"),
        Variable(name="disc1.x", value=np.array([1.0]), units=None),
        Variable(name="disc1.z", value=np.array([5.0, 2.0]), units="m**2"),
        Variable(name="disc1.y1", value=np.array([1.0]), units=None),
        Variable(name="disc1.y2", value=np.array([1.0]), units=None),
        Variable(name="disc2.z", value=np.array([5.0, 2.0]), units="m**2"),
        Variable(name="disc2.y1", value=np.array([1.0]), units=None),
        Variable(name="disc2.y2", value=np.array([1.0]), units=None),
        Variable(name="functions.x", value=np.array([1.0]), units=None),
        Variable(name="functions.z", value=np.array([5.0, 2.0]), units="m**2"),
        Variable(name="functions.y1", value=np.array([1.0]), units=None),
        Variable(name="functions.y2", value=np.array([1.0]), units=None),
        Variable(name="functions.g1", value=np.array([1.0]), units=None),
        Variable(name="functions.g2", value=np.array([1.0]), units=None),
        Variable(name="functions.f", value=np.array([1.0]), units=None),
    ]
    _test_and_check(problem, False, True, True, expected_computed_vars)

    # Check with the whole Sellar problem --------------------------------------
    # WITHOUT promotions, WITH computation
    expected_computed_vars = [
        Variable(name="indeps.x", value=np.array([1.0]), units="Pa"),
        Variable(name="indeps.z", value=np.array([5.0, 2.0]), units="m**2"),
        Variable(name="disc1.x", value=np.array([1.0]), units=None),
        Variable(name="disc1.z", value=np.array([5.0, 2.0]), units="m**2"),
        Variable(name="disc1.y1", value=np.array([25.58830237]), units=None),
        Variable(name="disc1.y2", value=np.array([12.05848815]), units=None),
        Variable(name="disc2.z", value=np.array([5.0, 2.0]), units="m**2"),
        Variable(name="disc2.y1", value=np.array([25.58830237]), units=None),
        Variable(name="disc2.y2", value=np.array([12.05848815]), units=None),
        Variable(name="functions.x", value=np.array([1.0]), units=None),
        Variable(name="functions.z", value=np.array([5.0, 2.0]), units="m**2"),
        Variable(name="functions.y1", value=np.array([25.58830237]), units=None),
        Variable(name="functions.y2", value=np.array([12.05848815]), units=None),
        Variable(name="functions.g1", value=np.array([-22.42830237]), units=None),
        Variable(name="functions.g2", value=np.array([-11.94151185]), units=None),
        Variable(name="functions.f", value=np.array([28.58830817]), units=None),
    ]
    problem.setup()
    problem.run_model()
    _test_and_check(problem, True, True, True, expected_vars)
    _test_and_check(problem, False, True, True, expected_computed_vars)


def test_variables_from_unconnected_inputs():
    def _test_and_check(
        problem: om.Problem,
        expected_mandatory_vars: List[Variable],
        expected_optional_vars: List[Variable],
    ):
        problem.setup()
        vars = VariableList.from_unconnected_inputs(problem, with_optional_inputs=False)
        assert vars == expected_mandatory_vars

        vars = VariableList.from_unconnected_inputs(problem, with_optional_inputs=True)
        assert vars == expected_mandatory_vars + expected_optional_vars

    # Check with an ExplicitComponent ------------------------------------------
    problem = om.Problem(Disc1())
    expected_mandatory_vars = [Variable(name="x", value=np.array([np.nan]), units=None)]
    expected_optional_vars = [
        Variable(name="z", value=np.array([5.0, 2.0]), units="m**2"),
        Variable(name="y2", value=np.array([1.0]), units=None),
    ]
    _test_and_check(problem, expected_mandatory_vars, expected_optional_vars)

    # Check with a Group -------------------------------------------------------
    group = om.Group()
    group.add_subsystem("disc1", Disc1(), promotes=["*"])
    group.add_subsystem("disc2", Disc2(), promotes=["*"])
    problem = om.Problem(group)

    expected_mandatory_vars = [Variable(name="x", value=np.array([np.nan]), units=None)]
    expected_optional_vars = [Variable(name="z", value=np.array([5.0, 2.0]), units="m**2")]
    _test_and_check(problem, expected_mandatory_vars, expected_optional_vars)

    # Check with the whole Sellar problem --------------------------------------
    # 'z' variable should now be mandatory, because it is so in Functions
    group = om.Group()
    group.add_subsystem("disc1", Disc1(), promotes=["*"])
    group.add_subsystem("disc2", Disc2(), promotes=["*"])
    group.add_subsystem("functions", Functions(), promotes=["*"])
    problem = om.Problem(group)

    expected_mandatory_vars = [
        Variable(name="x", value=np.array([np.nan]), units=None),
        Variable(name="z", value=np.array([np.nan, np.nan]), units="m**2"),
    ]
    expected_optional_vars = []
    _test_and_check(problem, expected_mandatory_vars, expected_optional_vars)



"""
    FAST - Copyright (c) 2016 ONERA ISAE
"""

#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import math

import numpy as np
from fastoad.models.aerodynamics.constants import POLAR_POINT_COUNT
from openmdao.core.explicitcomponent import ExplicitComponent


class Cd0NacelleAndPylons(ExplicitComponent):
    def initialize(self):
        self.options.declare("low_speed_aero", default=False, types=bool)

    def setup(self):
        self.low_speed_aero = self.options["low_speed_aero"]

        nans_array = np.full(POLAR_POINT_COUNT, np.nan)
        if self.low_speed_aero:
            self.add_input("reynolds_low_speed", val=np.nan)
            self.add_input("Mach_low_speed", val=np.nan)
            self.add_output("cd0_nacelle_low_speed")
            self.add_output("cd0_pylon_low_speed")
        else:
            self.add_input("data:aerodynamics:wing:cruise:reynolds", val=np.nan)
            self.add_input("data:aerodynamics:aircraft:cruise:CL", val=nans_array)
            self.add_input("data:TLAR:cruise_mach", val=np.nan)
            self.add_output("data:aerodynamics:nacelles:cruise:CD0")
            self.add_output("data:aerodynamics:pylons:cruise:CD0")

        self.add_input("data:geometry:propulsion:pylon:length", val=np.nan, units="m")
        self.add_input("data:geometry:propulsion:nacelle:length", val=np.nan, units="m")
        self.add_input("data:geometry:propulsion:pylon:wetted_area", val=np.nan, units="m**2")
        self.add_input("data:geometry:propulsion:nacelle:wetted_area", val=np.nan, units="m**2")
        self.add_input("data:geometry:propulsion:engine:count", val=np.nan)
        self.add_input("data:geometry:propulsion:fan:length", val=np.nan, units="m")
        self.add_input("data:geometry:wing:area", val=np.nan, units="m**2")

        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):
        pylon_length = inputs["data:geometry:propulsion:pylon:length"]
        nac_length = inputs["data:geometry:propulsion:nacelle:length"]
        wet_area_pylon = inputs["data:geometry:propulsion:pylon:wetted_area"]
        wet_area_nac = inputs["data:geometry:propulsion:nacelle:wetted_area"]
        n_engines = inputs["data:geometry:propulsion:engine:count"]
        fan_length = inputs["data:geometry:propulsion:fan:length"]
        wing_area = inputs["data:geometry:wing:area"]
        if self.low_speed_aero:
            mach = inputs["Mach_low_speed"]
            re_hs = inputs["reynolds_low_speed"]
        else:
            mach = inputs["data:TLAR:cruise_mach"]
            re_hs = inputs["data:aerodynamics:wing:cruise:reynolds"]

        cf_pylon_hs = 0.455 / (
            (1 + 0.144 * mach ** 2) ** 0.65 * (math.log10(re_hs * pylon_length)) ** 2.58
        )
        cf_nac_hs = 0.455 / (
            (1 + 0.144 * mach ** 2) ** 0.65 * (math.log10(re_hs * nac_length)) ** 2.58
        )

        # cd0 Pylon
        el_pylon = 0.06
        ke_cd0_pylon = 4.688 * el_pylon ** 2 + 3.146 * el_pylon
        cd0_pylon_hs = n_engines * (1 + ke_cd0_pylon) * cf_pylon_hs * wet_area_pylon / wing_area

        # cd0 Nacelles
        e_fan = 0.22
        kn_cd0_nac = 1 + 0.05 + 5.8 * e_fan / fan_length
        cd0_int_nac = 0.0002
        cd0_nac_hs = n_engines * (kn_cd0_nac * cf_nac_hs * wet_area_nac / wing_area + cd0_int_nac)

        if self.low_speed_aero:
            outputs["cd0_pylon_low_speed"] = cd0_pylon_hs
            outputs["cd0_nacelle_low_speed"] = cd0_nac_hs
        else:
            outputs["data:aerodynamics:pylons:cruise:CD0"] = cd0_pylon_hs
            outputs["data:aerodynamics:nacelles:cruise:CD0"] = cd0_nac_hs



"""
This module is for registering all internal OpenMDAO modules that we want
available through OpenMDAOSystemRegistry
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from fastoad.models.aerodynamics.aerodynamics_high_speed import AerodynamicsHighSpeed
from fastoad.models.aerodynamics.aerodynamics_landing import AerodynamicsLanding
from fastoad.models.geometry import Geometry
from fastoad.models.handling_qualities.compute_static_margin import ComputeStaticMargin
from fastoad.models.handling_qualities.tail_sizing.compute_tail_areas import ComputeTailAreas
from fastoad.models.loops.compute_wing_area import ComputeWingArea
from fastoad.models.performances import BreguetFromOWE
from fastoad.models.propulsion.fuel_engine.rubber_engine import OMRubberEngine
from fastoad.models.weight.weight import Weight
from fastoad.module_management import OpenMDAOSystemRegistry
from fastoad.module_management.constants import ModelDomain


def register_openmdao_systems():
    """
    The place where to register FAST-OAD internal models.

    Warning: this function is effective only if called from a Python module that
    is a started bundle for iPOPO
    """
    # Aerodynamics ################################################################
    OpenMDAOSystemRegistry.register_system(
        AerodynamicsLanding, "fastoad.aerodynamics.landing.legacy", domain=ModelDomain.AERODYNAMICS
    )
    OpenMDAOSystemRegistry.register_system(
        AerodynamicsHighSpeed,
        "fastoad.aerodynamics.highspeed.legacy",
        domain=ModelDomain.AERODYNAMICS,
    )

    # Geometry ####################################################################
    OpenMDAOSystemRegistry.register_system(
        Geometry, "fastoad.geometry.legacy", domain=ModelDomain.GEOMETRY
    )

    # handling qualities ##########################################################
    OpenMDAOSystemRegistry.register_system(
        ComputeTailAreas,
        "fastoad.handling_qualities.tail_sizing",
        domain=ModelDomain.HANDLING_QUALITIES,
    )
    OpenMDAOSystemRegistry.register_system(
        ComputeStaticMargin,
        "fastoad.handling_qualities.static_margin",
        domain=ModelDomain.HANDLING_QUALITIES,
    )

    # Loops #######################################################################
    OpenMDAOSystemRegistry.register_system(
        ComputeWingArea, "fastoad.loop.wing_area", domain=ModelDomain.OTHER
    )

    # Weight ######################################################################
    OpenMDAOSystemRegistry.register_system(
        Weight, "fastoad.weight.legacy", domain=ModelDomain.WEIGHT
    )
    # Performance #################################################################
    OpenMDAOSystemRegistry.register_system(
        BreguetFromOWE, "fastoad.performances.breguet", domain=ModelDomain.PERFORMANCE
    )

    # Propulsion ##################################################################
    rubber_engine_description = """
    Parametric engine model as OpenMDAO component.
    
    Implementation of E. Roux models for fuel consumption of low bypass ratio engines
    For more information, see RubberEngine class in FAST-OAD developer documentation.
    """

    OpenMDAOSystemRegistry.register_system(
        OMRubberEngine,
        "fastoad.propulsion.rubber_engine",
        domain=ModelDomain.PROPULSION,
        desc=rubber_engine_description,
    )



"""
Basic "Hello World" services using iPOPO decorators
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Dummy classes, so let's make pylint not too picky
# pylint: disable=missing-docstring
# pylint: disable=no-self-use
# pylint: disable=too-few-public-methods

from pelix.ipopo.decorators import ComponentFactory, Provides, Instantiate, Property

# Define the component factory, with a given name


# First, let's register 2 factories the iPOPO way
@ComponentFactory("hello-world-factory")
@Provides("hello.world")
@Property("_Prop1", "Prop1", 1)
@Property("_Prop_2", "Prop 2", "Says.Hello")
@Property("Instantiated", None, True)
@Instantiate("provider")
class Greetings1:
    def hello(self, name="World"):
        return "Hello, {0}!".format(name)


@ComponentFactory("hello-world-factory2")
@Provides("hello.world")
@Property("_Prop1", "Prop1", 2)
@Property("_Prop_2", "Prop 2", "Says.Hi")
@Property("Instantiated", None, True)
@Instantiate("provider2")
class Greetings2:
    def hello(self, name="World"):
        return "Hi, {0}!".format(name)



"""
Simple module for performances
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from fastoad.constants import FlightPhase
from fastoad.utils.physics import Atmosphere
from scipy.constants import g

CLIMB_MASS_RATIO = 0.97  # = mass at end of climb / mass at start of climb
DESCENT_MASS_RATIO = 0.98  # = mass at end of descent / mass at start of descent
RESERVE_MASS_RATIO = 0.06  # = (weight of fuel reserve)/ZFW
CLIMB_DESCENT_DISTANCE = 500  # in km, distance of climb + descent


class BreguetFromMTOW(om.Group):
    """
    Estimation of fuel consumption through Breguet formula with a rough estimate
    of climb and descent phases.

    MTOW (Max TakeOff Weight) being an input, the model computes the ZFW (Zero Fuel
    Weight) considering that all fuel but the reserve has been consumed during the
    mission.
    This model does not ensure consistency with OWE (Operating Empty Weight)
    """

    # TODO: in a more general case, this module will link the starting mass to
    #   the ending mass. Could we make the module more generic ?
    def setup(self):
        self.add_subsystem("propulsion", _BreguetPropulsion(), promotes=["*"])
        self.add_subsystem("distances", _Distances(), promotes=["*"])
        self.add_subsystem("cruise_mass_ratio", _CruiseMassRatio(), promotes=["*"])
        self.add_subsystem("fuel_weights", _FuelWeightFromMTOW(), promotes=["*"])
        self.add_subsystem("consumption", _Consumption(), promotes=["*"])


class BreguetFromOWE(om.Group):
    """
    Estimation of fuel consumption through Breguet formula with a rough estimate
    of climb and descent phases.

    For the sizing mission, the Breguet formula links MTOW (Max TakeOff Weight) to
    ZFW (Zero Fuel Weight).
    OWE (Operating Weight Empty) being linked to ZFW and MTOW, a cycle is implemented
    to have consistency between these 3 values.
    """

    def setup(self):
        self.add_subsystem("propulsion", _BreguetPropulsion(), promotes=["*"])
        self.add_subsystem("distances", _Distances(), promotes=["*"])
        self.add_subsystem("cruise_mass_ratio", _CruiseMassRatio(), promotes=["*"])
        self.add_subsystem("mtow", _MTOWFromOWE(), promotes=["*"])
        self.add_subsystem("fuel_weights", _FuelWeightFromMTOW(), promotes=["*"])
        self.add_subsystem("consumption", _Consumption(), promotes=["*"])

        self.nonlinear_solver = om.NewtonSolver()
        self.nonlinear_solver.options["iprint"] = 0
        self.nonlinear_solver.options["solve_subsystems"] = False
        self.nonlinear_solver.linesearch = om.BoundsEnforceLS()
        self.linear_solver = om.DirectSolver()


class _Consumption(om.ExplicitComponent):
    """
    Adds a variable for consumption/passenger/km
    """

    def setup(self):
        self.add_input("data:mission:sizing:fuel", np.nan, units="kg")
        self.add_input("data:TLAR:range", np.nan, units="km")
        self.add_input("data:TLAR:NPAX", np.nan)

        self.add_output("data:mission:sizing:fuel:unitary", units="kg/km")
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        fuel = inputs["data:mission:sizing:fuel"]
        npax = inputs["data:TLAR:NPAX"]
        distance = inputs["data:TLAR:range"]

        outputs["data:mission:sizing:fuel:unitary"] = fuel / npax / distance


class _BreguetPropulsion(om.ExplicitComponent):
    """
    Link with engine computation
    """

    def setup(self):
        self.add_input("data:mission:sizing:cruise:altitude", np.nan, units="m")
        self.add_input("data:TLAR:cruise_mach", np.nan)
        self.add_input("data:weight:aircraft:MTOW", np.nan, units="kg")
        self.add_input("data:aerodynamics:aircraft:cruise:L_D_max", np.nan)
        self.add_input("data:geometry:propulsion:engine:count", 2)

        self.add_output("data:propulsion:phase", FlightPhase.CRUISE)
        self.add_output("data:propulsion:use_thrust_rate", False)
        self.add_output("data:propulsion:required_thrust_rate", 0.0, lower=0.0, upper=1.0)
        self.add_output("data:propulsion:required_thrust", units="N", ref=1e5)
        self.add_output("data:propulsion:altitude", units="m", ref=1e4)
        self.add_output("data:propulsion:mach")

        self.declare_partials("data:propulsion:phase", "*", method="fd")
        self.declare_partials("data:propulsion:use_thrust_rate", "*", method="fd")
        self.declare_partials("data:propulsion:required_thrust_rate", "*", method="fd")
        self.declare_partials("data:propulsion:required_thrust", "*", method="fd")
        self.declare_partials(
            "data:propulsion:altitude", "data:mission:sizing:cruise:altitude", method="fd"
        )
        self.declare_partials("data:propulsion:mach", "data:TLAR:cruise_mach", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        engine_count = inputs["data:geometry:propulsion:engine:count"]
        ld_ratio = inputs["data:aerodynamics:aircraft:cruise:L_D_max"]
        mtow = inputs["data:weight:aircraft:MTOW"]
        initial_cruise_mass = mtow * CLIMB_MASS_RATIO

        # Variables for propulsion
        outputs["data:propulsion:altitude"] = inputs["data:mission:sizing:cruise:altitude"]
        outputs["data:propulsion:mach"] = inputs["data:TLAR:cruise_mach"]

        outputs["data:propulsion:required_thrust"] = (
            initial_cruise_mass / ld_ratio * g / engine_count
        )


class _FuelWeightFromMTOW(om.ExplicitComponent):
    """
    Estimation of fuel consumption through Breguet formula with a rough estimate
    of climb and descent phases
    """

    def setup(self):
        self.add_input("data:TLAR:cruise_mach", np.nan)
        self.add_input("data:TLAR:range", np.nan, units="m")
        self.add_input("data:aerodynamics:aircraft:cruise:L_D_max", np.nan)
        self.add_input("data:propulsion:SFC", np.nan, units="kg/N/s")
        self.add_input("data:weight:aircraft:MTOW", np.nan, units="kg")
        self.add_input("data:mission:sizing:cruise:mass_ratio", np.nan)
        self.add_input("data:mission:sizing:cruise:altitude", np.nan, units="m")

        self.add_output("data:mission:sizing:ZFW", units="kg", ref=1e4)
        self.add_output("data:mission:sizing:fuel", units="kg", ref=1e4)
        self.add_output("data:mission:sizing:trip:fuel", units="kg", ref=1e4)
        self.add_output("data:mission:sizing:climb:fuel", units="kg", ref=1e4)
        self.add_output("data:mission:sizing:cruise:fuel", units="kg", ref=1e4)
        self.add_output("data:mission:sizing:descent:fuel", units="kg", ref=1e4)
        self.add_output("data:mission:sizing:fuel_reserve", units="kg", ref=1e4)

        self.declare_partials("data:mission:sizing:ZFW", "*", method="fd")
        self.declare_partials("data:mission:sizing:fuel", "*", method="fd")
        self.declare_partials("data:mission:sizing:trip:fuel", "*", method="fd")
        self.declare_partials("data:mission:sizing:climb:fuel", "*", method="fd")
        self.declare_partials("data:mission:sizing:cruise:fuel", "*", method="fd")
        self.declare_partials("data:mission:sizing:descent:fuel", "*", method="fd")
        self.declare_partials("data:mission:sizing:fuel_reserve", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        mtow = inputs["data:weight:aircraft:MTOW"]
        cruise_mass_ratio = inputs["data:mission:sizing:cruise:mass_ratio"]

        flight_mass_ratio = cruise_mass_ratio * CLIMB_MASS_RATIO * DESCENT_MASS_RATIO
        zfw = mtow * flight_mass_ratio / (1.0 + RESERVE_MASS_RATIO)
        mission_fuel = mtow - zfw

        outputs["data:mission:sizing:ZFW"] = zfw

        outputs["data:mission:sizing:fuel"] = mission_fuel
        outputs["data:mission:sizing:trip:fuel"] = mtow * (1.0 - flight_mass_ratio)
        outputs["data:mission:sizing:climb:fuel"] = mtow * (1.0 - CLIMB_MASS_RATIO)
        outputs["data:mission:sizing:cruise:fuel"] = (
            mtow * CLIMB_MASS_RATIO * (1.0 - cruise_mass_ratio)
        )
        outputs["data:mission:sizing:descent:fuel"] = (
            mtow * CLIMB_MASS_RATIO * cruise_mass_ratio * (1.0 - DESCENT_MASS_RATIO)
        )
        outputs["data:mission:sizing:fuel_reserve"] = zfw * RESERVE_MASS_RATIO


class _MTOWFromOWE(om.ImplicitComponent):
    """
    Estimation of fuel consumption through Breguet formula with a rough estimate
    of climb and descent phases
    """

    def setup(self):
        self.add_input("data:mission:sizing:cruise:mass_ratio", np.nan)
        self.add_input("data:TLAR:range", np.nan, units="m")
        self.add_input("data:weight:aircraft:OWE", np.nan, units="kg")
        self.add_input("data:weight:aircraft:payload", np.nan, units="kg")

        self.add_output("data:weight:aircraft:MTOW", units="kg", ref=1e5)

        self.declare_partials("data:weight:aircraft:MTOW", "*", method="fd")

    def apply_nonlinear(self, inputs, outputs, residuals):
        owe = inputs["data:weight:aircraft:OWE"]
        payload_weight = inputs["data:weight:aircraft:payload"]
        cruise_mass_ratio = inputs["data:mission:sizing:cruise:mass_ratio"]

        mtow = outputs["data:weight:aircraft:MTOW"]

        flight_mass_ratio = cruise_mass_ratio * CLIMB_MASS_RATIO * DESCENT_MASS_RATIO
        zfw = mtow * flight_mass_ratio / (1.0 + RESERVE_MASS_RATIO)
        mission_owe = zfw - payload_weight

        residuals["data:weight:aircraft:MTOW"] = owe - mission_owe

    def guess_nonlinear(
        self, inputs, outputs, residuals, discrete_inputs=None, discrete_outputs=None
    ):
        # pylint: disable=too-many-arguments # It's OpenMDAO's fault :)
        outputs["data:weight:aircraft:MTOW"] = inputs["data:weight:aircraft:OWE"] * 1.5


class _Distances(om.ExplicitComponent):
    """ Rough estimation of distances for each flight phase"""

    def setup(self):
        self.add_input("data:TLAR:range", np.nan, units="m")

        self.add_output("data:mission:sizing:climb:distance", units="m", ref=1e3)
        self.add_output("data:mission:sizing:cruise:distance", units="m", ref=1e3)
        self.add_output("data:mission:sizing:descent:distance", units="m", ref=1e3)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        flight_range = inputs["data:TLAR:range"]

        outputs["data:mission:sizing:cruise:distance"] = (
            flight_range - CLIMB_DESCENT_DISTANCE * 1000.0
        )
        outputs["data:mission:sizing:climb:distance"] = CLIMB_DESCENT_DISTANCE * 500.0
        outputs["data:mission:sizing:descent:distance"] = CLIMB_DESCENT_DISTANCE * 500.0


class _CruiseMassRatio(om.ExplicitComponent):
    """
    Estimation of fuel consumption through Breguet formula for a given cruise distance
    """

    def setup(self):
        self.add_input("data:aerodynamics:aircraft:cruise:L_D_max", np.nan)
        self.add_input("data:propulsion:SFC", np.nan, units="kg/N/s")
        self.add_input("data:TLAR:cruise_mach", np.nan)
        self.add_input("data:mission:sizing:cruise:altitude", np.nan, units="m")
        self.add_input("data:mission:sizing:cruise:distance", np.nan, units="m")

        self.add_output("data:mission:sizing:cruise:mass_ratio")

        self.declare_partials("data:mission:sizing:cruise:mass_ratio", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        atmosphere = Atmosphere(
            inputs["data:mission:sizing:cruise:altitude"], altitude_in_feet=False
        )
        cruise_speed = atmosphere.speed_of_sound * inputs["data:TLAR:cruise_mach"]

        cruise_distance = inputs["data:mission:sizing:cruise:distance"]
        ld_ratio = inputs["data:aerodynamics:aircraft:cruise:L_D_max"]
        sfc = inputs["data:propulsion:SFC"]

        range_factor = cruise_speed * ld_ratio / g / sfc
        # During first iterations, SFC will be incorrect and range_factor may be too low,
        # resulting in null or too small cruise_mass_ratio.
        # Forcing cruise_mass_ratio to a minimum of 0.3 avoids problems and should not
        # harm (no airplane loses 70% of its weight from fuel consumption)
        cruise_mass_ratio = np.maximum(0.3, 1.0 / np.exp(cruise_distance / range_factor))

        outputs["data:mission:sizing:cruise:mass_ratio"] = cruise_mass_ratio



"""
    FAST - Copyright (c) 2016 ONERA ISAE
"""

#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from fastoad.models.aerodynamics.components.cd0_fuselage import Cd0Fuselage
from fastoad.models.aerodynamics.components.cd0_ht import Cd0HorizontalTail
from fastoad.models.aerodynamics.components.cd0_nacelle_pylons import Cd0NacelleAndPylons
from fastoad.models.aerodynamics.components.cd0_total import Cd0Total
from fastoad.models.aerodynamics.components.cd0_vt import Cd0VerticalTail
from fastoad.models.aerodynamics.components.cd0_wing import Cd0Wing
from fastoad.models.aerodynamics.components.cd_compressibility import CdCompressibility
from fastoad.models.aerodynamics.components.cd_trim import CdTrim
from fastoad.models.aerodynamics.components.compute_low_speed_aero import (
    ComputeAerodynamicsLowSpeed,
)
from fastoad.models.aerodynamics.components.compute_polar import ComputePolar
from fastoad.models.aerodynamics.components.compute_reynolds import ComputeReynolds
from fastoad.models.aerodynamics.components.initialize_cl import InitializeClPolar
from fastoad.models.aerodynamics.components.oswald import OswaldCoefficient
from openmdao.core.group import Group
from openmdao.core.indepvarcomp import IndepVarComp


class AerodynamicsLowSpeed(Group):
    """
    Models for low speed aerodynamics
    """

    def setup(self):
        self.add_subsystem("compute_low_speed_aero", ComputeAerodynamicsLowSpeed(), promotes=["*"])
        ivc = IndepVarComp("Mach_low_speed", val=0.2)
        self.add_subsystem("mach_low_speed", ivc, promotes=["*"])
        self.add_subsystem(
            "compute_oswald_coeff", OswaldCoefficient(low_speed_aero=True), promotes=["*"]
        )
        self.add_subsystem("comp_re", ComputeReynolds(low_speed_aero=True), promotes=["*"])
        self.add_subsystem("initialize_cl", InitializeClPolar(low_speed_aero=True), promotes=["*"])
        self.add_subsystem("cd0_wing", Cd0Wing(low_speed_aero=True), promotes=["*"])
        self.add_subsystem("cd0_fuselage", Cd0Fuselage(low_speed_aero=True), promotes=["*"])
        self.add_subsystem("cd0_ht", Cd0HorizontalTail(low_speed_aero=True), promotes=["*"])
        self.add_subsystem("cd0_vt", Cd0VerticalTail(low_speed_aero=True), promotes=["*"])
        self.add_subsystem(
            "cd0_nac_pylons", Cd0NacelleAndPylons(low_speed_aero=True), promotes=["*"]
        )
        self.add_subsystem("cd0_total", Cd0Total(low_speed_aero=True), promotes=["*"])
        self.add_subsystem("cd_comp", CdCompressibility(low_speed_aero=True), promotes=["*"])
        self.add_subsystem("cd_trim", CdTrim(low_speed_aero=True), promotes=["*"])
        self.add_subsystem("get_polar", ComputePolar(low_speed_aero=True), promotes=["*"])



"""
Aero computation for landing phase
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om

from fastoad.models.aerodynamics.components.compute_max_cl_landing import ComputeMaxClLanding
from fastoad.models.aerodynamics.components.high_lift_aero import ComputeDeltaHighLift
from fastoad.models.aerodynamics.external.xfoil import XfoilPolar
from fastoad.models.aerodynamics.external.xfoil.xfoil_polar import OPTION_XFOIL_EXE_PATH
from fastoad.models.options import OpenMdaoOptionDispatcherGroup
from fastoad.utils.physics import Atmosphere


class AerodynamicsLanding(OpenMdaoOptionDispatcherGroup):
    """
    Computes maximum CL of the aircraft in landing conditions.

    Maximum 2D CL without high-lift is computed using XFoil (or provided as input if option
    use_xfoil is set to False). 3D CL is deduced using sweep angle.

    Contribution of high-lift devices is modelled according to their geometry (span and chord ratio)
    and their deflection angles.

    Options:
      - use_xfoil:
         - if True, maximum 2D CL without high-lift aerodynamics:aircraft:landing:CL_max_clean_2D
           is computed using XFOIL
         - if False, aerodynamics:aircraft:landing:CL_max_clean_2D must be provided as input (but
           process is faster)
      - alpha_min, alpha_max:
         - used if use_xfoil is True. Sets the alpha range that is explored to find maximum 2D CL
           without high-lift
      - xfoil_exe_path:
         - the path to the XFOIL executable. Needed for non-Windows OS.
    """

    def initialize(self):
        self.options.declare("use_xfoil", default=True, types=bool)
        self.options.declare("xfoil_alpha_min", default=0.0, types=float)
        self.options.declare("xfoil_alpha_max", default=30.0, types=float)
        self.options.declare("xfoil_iter_limit", default=500, types=int)
        self.options.declare(OPTION_XFOIL_EXE_PATH, default="", types=str, allow_none=True)

    def setup(self):
        self.add_subsystem("mach_reynolds", ComputeMachReynolds(), promotes=["*"])
        if self.options["use_xfoil"]:
            start = self.options["xfoil_alpha_min"]
            end = self.options["xfoil_alpha_max"]
            iter_limit = self.options["xfoil_iter_limit"]
            self.add_subsystem(
                "xfoil_run",
                XfoilPolar(alpha_start=start, alpha_end=end, iter_limit=iter_limit),
                promotes=["data:geometry:wing:thickness_ratio"],
            )
        self.add_subsystem("CL_2D_to_3D", Compute3DMaxCL(), promotes=["*"])
        self.add_subsystem(
            "delta_cl_landing", ComputeDeltaHighLift(landing_flag=True), promotes=["*"]
        )
        self.add_subsystem("compute_max_cl_landing", ComputeMaxClLanding(), promotes=["*"])

        if self.options["use_xfoil"]:
            self.connect("data:aerodynamics:aircraft:landing:mach", "xfoil_run.xfoil:mach")
            self.connect("data:aerodynamics:aircraft:landing:reynolds", "xfoil_run.xfoil:reynolds")
            self.connect(
                "xfoil_run.xfoil:CL_max_2D", "data:aerodynamics:aircraft:landing:CL_max_clean_2D"
            )


class ComputeMachReynolds(om.ExplicitComponent):
    """
    Mach and Reynolds computation
    """

    def setup(self):
        self.add_input("data:geometry:wing:MAC:length", val=np.nan, units="m")
        self.add_input("data:TLAR:approach_speed", val=np.nan, units="m/s")
        self.add_output("data:aerodynamics:aircraft:landing:mach")
        self.add_output("data:aerodynamics:aircraft:landing:reynolds")

        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        l0_wing = inputs["data:geometry:wing:MAC:length"]
        speed = inputs["data:TLAR:approach_speed"]

        atm = Atmosphere(0.0, 15.0)
        mach = speed / atm.speed_of_sound
        reynolds = atm.get_unitary_reynolds(mach) * l0_wing

        outputs["data:aerodynamics:aircraft:landing:mach"] = mach
        outputs["data:aerodynamics:aircraft:landing:reynolds"] = reynolds


class Compute3DMaxCL(om.ExplicitComponent):
    """
    Computes 3D max CL from 2D CL (XFOIL-computed) and sweep angle
    """

    def setup(self):
        self.add_input("data:geometry:wing:sweep_25", val=np.nan, units="rad")
        self.add_input("data:aerodynamics:aircraft:landing:CL_max_clean_2D", val=np.nan)

        self.add_output("data:aerodynamics:aircraft:landing:CL_max_clean")

        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        sweep_25 = inputs["data:geometry:wing:sweep_25"]
        cl_max_2d = inputs["data:aerodynamics:aircraft:landing:CL_max_clean_2D"]
        outputs["data:aerodynamics:aircraft:landing:CL_max_clean"] = (
            cl_max_2d * 0.9 * np.cos(sweep_25)
        )



"""
Computation of wing area
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openmdao.api as om
from scipy.constants import g


class ComputeWingArea(om.Group):
    """
    Computes needed wing area for:
      - having enough lift at required approach speed
      - being able to load enough fuel to achieve the sizing mission
    """

    def setup(self):
        self.add_subsystem("wing_area", _ComputeWingArea(), promotes=["*"])
        self.add_subsystem("constraints", _ComputeWingAreaConstraints(), promotes=["*"])


class _ComputeWingArea(om.ExplicitComponent):
    """ Computation of wing area from needed approach speed and mission fuel """

    def setup(self):
        self.add_input("data:geometry:wing:aspect_ratio", val=np.nan)
        self.add_input("data:geometry:wing:root:thickness_ratio", val=np.nan)
        self.add_input("data:geometry:wing:tip:thickness_ratio", val=np.nan)
        self.add_input("data:mission:sizing:fuel", val=np.nan, units="kg")
        self.add_input("data:TLAR:approach_speed", val=np.nan, units="m/s")

        self.add_input("data:weight:aircraft:MLW", val=np.nan, units="kg")
        self.add_input("data:aerodynamics:aircraft:landing:CL_max", val=np.nan)

        self.add_output("data:geometry:wing:area", val=10.0, units="m**2")
        self.declare_partials("data:geometry:wing:area", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        lambda_wing = inputs["data:geometry:wing:aspect_ratio"]
        root_thickness_ratio = inputs["data:geometry:wing:root:thickness_ratio"]
        tip_thickness_ratio = inputs["data:geometry:wing:tip:thickness_ratio"]
        mfw_mission = inputs["data:mission:sizing:fuel"]
        wing_area_mission = (
            (mfw_mission - 1570.0)
            / 224
            / lambda_wing ** -0.4
            / (0.6 * root_thickness_ratio + 0.4 * tip_thickness_ratio)
        ) ** (1.0 / 1.5)

        approach_speed = inputs["data:TLAR:approach_speed"]
        mlw = inputs["data:weight:aircraft:MLW"]
        max_CL = inputs["data:aerodynamics:aircraft:landing:CL_max"]
        wing_area_approach = 2 * mlw * g / ((approach_speed / 1.23) ** 2) / (1.225 * max_CL)

        outputs["data:geometry:wing:area"] = max(wing_area_mission, wing_area_approach)


class _ComputeWingAreaConstraints(om.ExplicitComponent):
    def setup(self):
        self.add_input("data:mission:sizing:fuel", val=np.nan, units="kg")
        self.add_input("data:weight:aircraft:MFW", val=np.nan, units="kg")

        self.add_input("data:TLAR:approach_speed", val=np.nan, units="m/s")
        self.add_input("data:weight:aircraft:MLW", val=np.nan, units="kg")
        self.add_input("data:aerodynamics:aircraft:landing:CL_max", val=np.nan)
        self.add_input("data:geometry:wing:area", val=np.nan, units="m**2")

        self.add_output("data:weight:aircraft:additional_fuel_capacity", units="kg")
        self.add_output("data:aerodynamics:aircraft:landing:additional_CL_capacity")

        self.declare_partials(
            "data:weight:aircraft:additional_fuel_capacity",
            ["data:weight:aircraft:MFW", "data:mission:sizing:fuel"],
            method="fd",
        )
        self.declare_partials(
            "data:aerodynamics:aircraft:landing:additional_CL_capacity",
            [
                "data:TLAR:approach_speed",
                "data:weight:aircraft:MLW",
                "data:aerodynamics:aircraft:landing:CL_max",
                "data:geometry:wing:area",
            ],
            method="fd",
        )

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        mfw = inputs["data:weight:aircraft:MFW"]
        mission_fuel = inputs["data:mission:sizing:fuel"]
        v_approach = inputs["data:TLAR:approach_speed"]
        cl_max = inputs["data:aerodynamics:aircraft:landing:CL_max"]
        mlw = inputs["data:weight:aircraft:MLW"]
        wing_area = inputs["data:geometry:wing:area"]

        outputs["data:weight:aircraft:additional_fuel_capacity"] = mfw - mission_fuel
        outputs["data:aerodynamics:aircraft:landing:additional_CL_capacity"] = cl_max - mlw * g / (
            0.5 * 1.225 * (v_approach / 1.23) ** 2 * wing_area
        )



"""
Utility functions for OpenMDAO classes/instances
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from copy import deepcopy
from logging import Logger
from typing import Tuple, List

import numpy as np
import openmdao.api as om


# pylint: disable=protected-access #  needed for OpenMDAO introspection


def get_unconnected_input_names(
    problem: om.Problem, logger: Logger = None
) -> Tuple[List[str], List[str]]:
    """
    For provided OpenMDAO problem, looks for inputs that are connected to no output.

    Inputs that have numpy.nan as default value are considered as mandatory. Other ones are
    considered as optional.

    If a logger is provided, it will issue errors for the first category, and warnings for the
    second one.

    :param problem: OpenMDAO Problem or System instance to inspect
    :param logger: optional logger instance
    :return: tuple(list of missing mandatory inputs, list of missing optional inputs)
    """

    # Setup() is needed
    model = get_problem_after_setup(problem).model

    prom2abs: dict = model._var_allprocs_prom2abs_list["input"]
    connexions: dict = model._conn_global_abs_in2out

    mandatory_unconnected = []
    optional_unconnected = []

    for abs_names in prom2abs.values():
        # At each iteration, get absolute names that match one promoted name, or one
        # absolute name that has not been promoted.
        unconnected = [a for a in abs_names if a not in connexions or len(connexions[a]) == 0]
        if unconnected:
            for abs_name in abs_names:
                value = model._var_abs2meta[abs_name]["value"]
                if np.all(np.isnan(value)):
                    mandatory_unconnected.append(abs_name)
                else:
                    optional_unconnected.append(abs_name)

    if logger:
        if mandatory_unconnected:
            logger.error("Following inputs are required and not connected:")
            for abs_name in sorted(mandatory_unconnected):
                logger.error("    %s", abs_name)

        if optional_unconnected:
            logger.warning(
                "Following inputs are not connected so their default value will be used:"
            )
            for abs_name in sorted(optional_unconnected):
                value = model._var_abs2meta[abs_name]["value"]
                logger.warning("    %s : %s", abs_name, value)

    return mandatory_unconnected, optional_unconnected


def get_problem_after_setup(problem: om.Problem) -> om.Problem:
    """
    This method should be used when an operation is needed that requires setup() to be run, without
    having the problem being actually setup.

    :param problem:
    :return: the problem itself it setup() has already been run, or a copy of the provided problem after setup()
             has been run
    """

    if problem._setup_status == 0:
        # If setup() has not been done, we create a copy of the problem so we can work
        # on the model without doing setup() out of user notice
        tmp_problem = deepcopy(problem)
        tmp_problem.setup()
        return tmp_problem
    else:
        return problem



# -*- coding: utf-8 -*-
"""
  Sellar functions
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from openmdao.api import ExplicitComponent


class FunctionsBase(ExplicitComponent):
    """ An OpenMDAO base component to encapsulate Functions discipline """

    def initialize(self):
        self.options.declare("best_doctor", 10)

    def setup(self):
        self.add_input("x", val=2, desc="")
        self.add_input(
            "z", val=[np.nan, np.nan], desc="", units="m**2"
        )  # NaN as default for testing connexion check
        self.add_input("y1", val=1.0, desc="")
        self.add_input("y2", val=1.0, desc="")

        self.add_output("f", val=1.0, desc="")

        self.add_output("g1", val=1.0, desc="")

        self.add_output("g2", val=1.0, desc="")
        self.declare_partials("*", "*", method="fd")



"""
    Estimation of center of gravity for load case 2
"""

#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import numpy as np
from openmdao.core.explicitcomponent import ExplicitComponent


class ComputeCGLoadCase2(ExplicitComponent):
    # TODO: Document equations. Cite sources
    """ Center of gravity estimation for load case 2 """

    def setup(self):
        self.add_input("data:geometry:wing:MAC:length", val=np.nan, units="m")
        self.add_input("data:geometry:wing:MAC:at25percent:x", val=np.nan, units="m")
        self.add_input("data:weight:payload:PAX:CG:x", val=np.nan, units="m")
        self.add_input("data:weight:payload:rear_fret:CG:x", val=np.nan, units="m")
        self.add_input("data:weight:payload:front_fret:CG:x", val=np.nan, units="m")
        self.add_input("data:TLAR:NPAX", val=np.nan)
        self.add_input("data:weight:aircraft:MFW", val=np.nan, units="kg")
        self.add_input("data:weight:fuel_tank:CG:x", val=np.nan, units="m")
        self.add_input("data:weight:aircraft_empty:CG:x", val=np.nan, units="m")
        self.add_input("data:weight:aircraft_empty:mass", val=np.nan, units="m")

        self.add_output("data:weight:aircraft:load_case_2:CG:MAC_position")

        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):
        l0_wing = inputs["data:geometry:wing:MAC:length"]
        fa_length = inputs["data:geometry:wing:MAC:at25percent:x"]
        cg_pax = inputs["data:weight:payload:PAX:CG:x"]
        cg_rear_fret = inputs["data:weight:payload:rear_fret:CG:x"]
        cg_front_fret = inputs["data:weight:payload:front_fret:CG:x"]
        npax = inputs["data:TLAR:NPAX"]
        x_cg_plane_aft = inputs["data:weight:aircraft_empty:CG:x"]
        x_cg_plane_down = inputs["data:weight:aircraft_empty:mass"]
        x_cg_plane_up = x_cg_plane_aft * x_cg_plane_down
        mfw = inputs["data:weight:aircraft:MFW"]
        cg_tank = inputs["data:weight:fuel_tank:CG:x"]

        weight_pax = npax * 90.0
        weight_rear_fret = npax * 20
        weight_front_fret = npax * 20
        weight_pl = weight_pax + weight_rear_fret + weight_front_fret
        x_cg_pl_2 = (
            weight_pax * cg_pax
            + weight_rear_fret * cg_rear_fret
            + weight_front_fret * cg_front_fret
        ) / (weight_pl)
        x_cg_plane_pl_2 = (x_cg_plane_up + weight_pl * x_cg_pl_2 + mfw * cg_tank) / (
            x_cg_plane_down + weight_pl + mfw
        )  # forward
        cg_ratio_pl_2 = (x_cg_plane_pl_2 - fa_length + 0.25 * l0_wing) / l0_wing

        outputs["data:weight:aircraft:load_case_2:CG:MAC_position"] = cg_ratio_pl_2



"""
Readers for legacy XML format
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from fastoad.io.xml import VariableXmlBaseFormatter
from fastoad.io.xml.translator import VarXpathTranslator
from importlib_resources import open_text

from . import resources

CONVERSION_FILENAME_1 = "legacy1.txt"


class VariableLegacy1XmlFormatter(VariableXmlBaseFormatter):
    """
    Formatter for legacy XML format (version "1")
    """

    def __init__(self):
        translator = VarXpathTranslator()
        with open_text(resources, CONVERSION_FILENAME_1) as translation_table:
            translator.read_translation_table(translation_table)
        super().__init__(translator)

        self.xml_unit_attribute = "unit"
